import jax.numpy as jnp

from ..._problem import AbstractNonlinearEquations


class ARGLALE(AbstractNonlinearEquations):
    """ARGLALE problem implementation.

    Variable dimension full rank linear problem, a linear equation version.

    Source: Problem 32 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#80 (with different N and M)
    SIF input: Ph. Toint, Dec 1989.

    classification NLR2-AN-V-V

    This is a(n infeasible) linear feasibility problem
    N is the number of free variables
    M is the number of equations ( M.ge.N)
    """

    # Default parameters
    N: int = 200  # Number of variables
    M: int = 400  # Number of equations (M >= N)

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def constraint(self, y):
        """Return the equality constraints (residuals) for the nonlinear system."""
        residuals = self.residual(y)
        # Return (equality_constraints, inequality_constraints)
        # For nonlinear equations, we have only equality constraints
        return residuals, None

    def residual(self, y):
        """Compute the residuals for the system (vectorized)."""
        n = self.N
        m = self.M
        x = y  # Variables

        # Compute coefficients
        minus_two_over_m = -2.0 / m

        # Total sum of all variables
        total_sum = jnp.sum(x)

        # For the first N residuals, each residual i has:
        # G(i) = sum_{j=1}^{i-1} (-2/M) * x_j + (1-2/M) * x_i + sum_{j=i+1}^{N}
        #        (-2/M) * x_j - 1
        # = (-2/M) * [sum_{j=1}^{N} x_j - x_i] + (1-2/M) * x_i - 1
        # = (-2/M) * sum(x) + (-2/M) * (-x_i) + (1-2/M) * x_i - 1
        # = (-2/M) * sum(x) + (2/M) * x_i + x_i - (2/M) * x_i - 1
        # = (-2/M) * sum(x) + x_i - 1

        # First N residuals
        residuals_first_n = minus_two_over_m * total_sum + x - 1.0

        # Remaining M-N residuals (all identical)
        # G(i) = sum_{j=1}^{N} (-2/M) * x_j - 1 for i = N+1 to M
        remaining_value = minus_two_over_m * total_sum - 1.0
        residuals_remaining = jnp.full(m - n, remaining_value)

        # Concatenate all residuals
        residuals = jnp.concatenate([residuals_first_n, residuals_remaining])

        return residuals

    @property
    def y0(self):
        """Initial guess for variables."""
        return jnp.ones(self.N)

    @property
    def bounds(self):
        """Returns None as this problem has no bounds."""
        return None

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        # For nonlinear equations, the objective is always zero
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: This implementation requires human review and verification against
# another CUTEst interface
class EG1(AbstractUnconstrainedMinimisation):
    """The EG1 function.

    A simple nonlinear problem given as an example in Section 1.2.3 of
    the LANCELOT Manual.

    Source:
    A.R. Conn, N. Gould and Ph.L. Toint,
    "LANCELOT, A Fortran Package for Large-Scale Nonlinear Optimization
    (Release A)" Springer Verlag, 1992.

    SIF input: N. Gould and Ph. Toint, June 1994.

    Classification: OBR2-AY-3-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 3  # Problem has 3 variables

    def objective(self, y, args):
        del args
        x1, x2, x3 = y

        # GROUP1: x1^2
        f1 = x1**2

        # GROUP2: (x2*x3)^4
        f2 = (x2 * x3) ** 4

        # GROUP3: combination of sin(x2 + x1 + x3) and x1*x3
        f3_1 = x2 * jnp.sin(x2 + x1 + x3)
        f3_2 = x1 * x3

        return f1 + f2 + f3_1 + f3_2

    @property
    def y0(self):
        # Initial values from SIF file
        return jnp.array([1.0, 1.0, 1.0])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        return None

    @property
    def expected_objective_value(self):
        return None



import jax
import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractUnconstrainedMinimisation


class TRIGON1(AbstractUnconstrainedMinimisation):
    """TRIGON1 problem - SCIPY global optimization benchmark example Trigonometric01.

    Fit: y = sum_{j=1}^{n} cos(x_j) + i (cos(x_i) + sin(x_i)) + e

    Source: Problem from the SCIPY benchmark set
    https://github.com/scipy/scipy/tree/master/benchmarks/benchmarks/go_benchmark_functions

    SIF input: Nick Gould, Jan 2020

    Classification: SUR2-MN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    _n: int = 10

    @property
    def n(self):
        """Number of variables."""
        return self._n

    def objective(self, y, args):
        """Compute the sum of squares objective."""
        del args
        x = y
        n = self.n

        # Compute residuals F(i) for i = 1, ..., n
        def compute_residual(i):
            # F(i) = sum cos(x_j) + (i+1) * (cos(x_i) + sin(x_i)) - (n + i + 1)
            cos_sum = jnp.sum(jnp.cos(x))
            i_float = inexact_asarray(i)
            individual_term = (i_float + 1) * (jnp.cos(x[i]) + jnp.sin(x[i]))
            target = float(n) + i_float + 1
            return cos_sum + individual_term - target

        residuals = jax.vmap(compute_residual)(jnp.arange(n))
        return jnp.sum(residuals**2)

    @property
    def y0(self):
        """Initial guess."""
        return inexact_asarray(jnp.full(self.n, 0.1))

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution."""
        return None  # Not provided in SIF

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HS70(AbstractConstrainedMinimisation):
    """HS70 problem - Water flow routing problem.

    This problem arises in water flow routing.

    TODO: Human review needed - test failures. The implementation follows the SIF file
    but tests fail. Challenges include:
    - Complex Y1 and Y2 function definitions with multiple terms
    - Y1 uses (1-x3) while Y2 uses x3 in different places
    - The functions involve intricate combinations of power, exponential and other terms
    - Constraint involves a product element (x3*x4)

    Source: problem 70 incorrectly stated in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    SIF input: Nick Gould, August 1991, modified May 2024

    Classification: SQR2-MN-4-1
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self):
        """Number of variables."""
        return 4

    @property
    def m(self):
        """Number of constraints."""
        return 1  # One inequality constraint

    def objective(self, y, args):
        """Compute the sum of squares objective."""
        del args
        x = y

        # Data points
        c = jnp.array(
            [
                0.1,
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
                6.0,
                7.0,
                8.0,
                9.0,
                10.0,
                11.0,
                12.0,
                13.0,
                14.0,
                15.0,
                16.0,
                17.0,
                18.0,
            ]
        )

        y_data = jnp.array(
            [
                0.00189,
                0.1038,
                0.268,
                0.506,
                0.577,
                0.604,
                0.725,
                0.898,
                0.947,
                0.845,
                0.702,
                0.528,
                0.385,
                0.257,
                0.159,
                0.0869,
                0.0453,
                0.01509,
                0.00189,
            ]
        )

        # Compute Y1 and Y2 functions for each data point
        def compute_y1(x1, x2, x3, x4, ci):
            b = x3 + x4 * (1.0 - x3)
            ci_scaled = ci / 7.658
            p0 = 1.0 + 1.0 / (12.0 * x2)
            p1 = 1.0 / p0
            p2 = x3
            p3 = b**x2
            c4 = jnp.sqrt(1.0 / (2.0 * jnp.pi))
            p4 = c4 * jnp.sqrt(x2)
            p5 = (1.0 / ci_scaled) * (ci_scaled**x2)
            p6 = jnp.exp(x2 * (1.0 - ci_scaled * b))
            return p1 * p2 * p3 * p4 * p5 * p6

        def compute_y2(x1, x2, x3, x4, ci):
            b = x3 + x4 * (1.0 - x3)
            ci_scaled = ci / 7.658
            p0 = 1.0 + 1.0 / (12.0 * x1)
            p1 = 1.0 / p0
            p2 = 1.0 - x3  # This is different from Y1!
            p3 = (b / x4) ** x1  # This is also different from Y1!
            c4 = jnp.sqrt(1.0 / (2.0 * jnp.pi))
            p4 = c4 * jnp.sqrt(x1)
            p5 = (1.0 / ci_scaled) * (ci_scaled**x1)
            p6 = jnp.exp(x1 * (1.0 - ci_scaled * b / x4))  # Also adjusted for b/x4
            return p1 * p2 * p3 * p4 * p5 * p6

        # Compute residuals
        residuals = jnp.zeros(19)
        for i in range(19):
            y1_val = compute_y1(x[0], x[1], x[2], x[3], c[i])
            y2_val = compute_y2(x[0], x[1], x[2], x[3], c[i])
            residuals = residuals.at[i].set(y1_val + y2_val - y_data[i])

        # Objective is sum of squares
        return jnp.sum(residuals**2)

    def constraint(self, y):
        """Compute the constraints."""
        x = y

        # Constraint: x3 + x4 + x3*x4 >= 1.0 (written as g(x) >= 0)
        # This comes from the linear part (x3 + x4) plus the PROD element (x3*x4)
        # with coefficient -1
        ineq_constraint = x[2] + x[3] + x[2] * x[3] - 1.0

        # Return as inequality constraint (empty equality constraints)
        return jnp.array([]), jnp.array([ineq_constraint])

    def equality_constraints(self):
        """Specify which constraints are equalities."""
        return jnp.array([])  # No equality constraints

    @property
    def y0(self):
        """Initial guess."""
        return jnp.array([2.0, 4.0, 0.04, 2.0])

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def bounds(self):
        """Variable bounds."""
        lower = jnp.array([0.00001, 0.00001, 0.00001, 0.00001])
        upper = jnp.array([100.0, 100.0, 1.0, 100.0])
        return lower, upper

    @property
    def expected_result(self):
        """Expected optimal solution."""
        return jnp.array([12.27695, 4.631788, 0.3128625, 2.029290])

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(0.007498464)  # From SIF file solution comment



import equinox as eqx
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


# TODO: Human review needed
# Attempts made: Fixed dtype issues, analyzed Hessian computation
# Suspected issues: Cross-derivative H[2,3] differs significantly from pycutest
# Additional resources needed: Verify the exact Hessian computation for scaled L2 groups
class HIMMELBF(AbstractUnconstrainedMinimisation):
    """Himmelblau 4 variable data fitting problem.

    A 4 variables data fitting problems by Himmelblau.

    Source: problem 32 in
    D.H. Himmelblau,
    "Applied nonlinear programming",
    McGraw-Hill, New-York, 1972.

    See Buckley#76 (p. 66)

    SIF input: Ph. Toint, Dec 1989.

    classification SUR2-AN-4-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem data
    a_data: tuple[float, ...] = eqx.field(
        default=(0.0, 0.000428, 0.001000, 0.001610, 0.002090, 0.003480, 0.005250),
        init=False,
    )
    b_data: tuple[float, ...] = eqx.field(
        default=(7.391, 11.18, 16.44, 16.20, 22.20, 24.02, 31.32), init=False
    )

    def objective(self, y, args):
        """Compute the objective function (vectorized).

        The objective is a sum of squared ratios.
        """
        x1, x2, x3, x4 = y

        # Convert data to JAX arrays for vectorized operations
        a_vals = jnp.array(self.a_data, dtype=y.dtype)
        b_vals = jnp.array(self.b_data, dtype=y.dtype)

        # Vectorized element function HF
        u = x1 * x1 + a_vals * x2 * x2 + a_vals * a_vals * x3 * x3
        v = b_vals * (1.0 + a_vals * x4 * x4)
        f_i = u / v

        # Vectorized group evaluation with constant 1.0 and scale 0.0001
        g_i = (f_i - 1.0) ** 2

        # Sum of all groups with scale (SCALE in SIF means multiply by 1/scale)
        return jnp.sum(g_i) / 0.0001

    @property
    def y0(self):
        return jnp.array([2.7, 90.0, 1500.0, 10.0])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        return None

    @property
    def expected_objective_value(self):
        return jnp.array(318.572)



import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractConstrainedMinimisation


class LUKVLI18(AbstractConstrainedMinimisation):
    """LUKVLI18 - Chained modified HS53 problem.

    Problem 5.18 from Luksan and Vlcek test problems with inequality constraints.

    The objective is a chained modified HS53 function:
    f(x) = Σ[i=1 to (n-1)/4] [(x_{j+1} - x_{j+2})^4 + (x_{j+2} + x_{j+3} - 2)^2 +
                               (x_{j+4} - 1)^2 + (x_{j+5} - 1)^2]
    where j = 4(i-1), l = 4*div(k-1,3)

    Subject to inequality constraints:
    c_k(x) = x_{l+1}^2 + 3x_{l+2} ≤ 0, for k ≡ 1 (mod 3), 1 ≤ k ≤ n_C
    c_k(x) = x_{l+3}^2 + x_{l+4} - 2x_{l+5} ≤ 0, for k ≡ 2 (mod 3), 1 ≤ k ≤ n_C
    c_k(x) = x_{l+2}^2 - x_{l+5} ≤ 0, for k ≡ 0 (mod 3), 1 ≤ k ≤ n_C
    where n_C = 3(n-1)/4

    Starting point: x_i = 2 for i = 1, ..., n

    Source: L. Luksan and J. Vlcek,
    "Sparse and partially separable test problems for
    unconstrained and equality constrained optimization",
    Technical Report 767, Inst. Computer Science, Academy of Sciences
    of the Czech Republic, 182 07 Prague, Czech Republic, 1999


    Equality constraints changed to inequalities

    SIF input: Nick Gould, April 2001

    Classification: OOR2-AY-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 9997  # Default dimension, (n-1) must be divisible by 4

    def objective(self, y, args):
        del args
        n = len(y)
        # Chained modified HS53 function - vectorized
        num_groups = (n - 1) // 4
        if num_groups == 0 or n < 5:
            return jnp.array(0.0)

        # For each group i=1..num_groups, we have j = 4*(i-1)
        # We need y[j] through y[j+4]
        i = jnp.arange(num_groups)
        j = 4 * i  # j values in 0-based

        # Extract elements for all groups
        y_j = y[j]  # y[j]
        y_j1 = y[j + 1]  # y[j+1]
        y_j2 = y[j + 2]  # y[j+2]
        y_j3 = y[j + 3]  # y[j+3]
        y_j4 = y[j + 4]  # y[j+4]

        # Compute all terms at once
        terms = (
            (y_j - y_j1) ** 4
            + (y_j1 + y_j2 - 2) ** 2
            + (y_j3 - 1) ** 2
            + (y_j4 - 1) ** 2
        )

        return jnp.sum(terms)

    @property
    def y0(self):
        # Starting point: x_i = 2 for all i
        return inexact_asarray(jnp.full(self.n, 2.0))

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution pattern based on problem structure
        return None  # Unknown exact solution

    @property
    def expected_objective_value(self):
        return None  # Unknown exact objective value

    @property
    def bounds(self):
        return None

    def constraint(self, y):
        n = len(y)
        n_c = 3 * (n - 1) // 4

        if n_c == 0:
            return None, jnp.array([])

        # Compute l values for all k
        k_values = jnp.arange(1, n_c + 1)
        l_values = 4 * ((k_values - 1) // 3)

        # Extend y with zeros to safely access all indices
        # Maximum index needed is max(l_values) + 5
        extended_length = n + 20  # Add sufficient padding
        y_extended = jnp.zeros(extended_length)
        y_extended = y_extended.at[:n].set(y)

        # Type 1 constraints: k ≡ 1 (mod 3)
        # c_k = x_{l+1}^2 + 3x_{l+2}
        c1 = y_extended[l_values] ** 2 + 3 * y_extended[l_values + 1]

        # Type 2 constraints: k ≡ 2 (mod 3)
        # c_k = x_{l+3}^2 + x_{l+4} - 2x_{l+5}
        c2 = (
            y_extended[l_values + 2] ** 2
            + y_extended[l_values + 3]
            - 2 * y_extended[l_values + 4]
        )

        # Type 3 constraints: k ≡ 0 (mod 3)
        # c_k = x_{l+2}^2 - x_{l+5}
        c3 = y_extended[l_values + 1] ** 2 - y_extended[l_values + 4]

        # Select constraints based on k modulo 3
        k_mod3 = k_values % 3
        constraints = jnp.where(k_mod3 == 1, c1, jnp.where(k_mod3 == 2, c2, c3))

        return None, constraints



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class FLT(AbstractConstrainedMinimisation):
    """FLT - A troublesome problem for filter methods.

    A simple 2-variable, 2-constraint optimization problem designed to test
    filter methods. The problem demonstrates challenging behavior for
    sequential quadratic programming (SQP) filter methods.

    Variables: X1, X2
    Objective: minimize (1 - X2)^2
    Constraints:
        CON1: X1^2 = 0
        CON2: X1^3 = 0

    Source: R. Fletcher, S. Leyffer and Ph. L. Toint,
    "On the global convergence of a filter-SQP method",
    SIAM J. Optimization 13 (2002):44-59.

    SIF input: Nick Gould, May 2008.

    Classification: QOR2-AN-2-2
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def y0(self):
        # Starting point from SIF file
        return jnp.array([1.0, 0.0])

    @property
    def args(self):
        return None

    def objective(self, y, args):
        del args
        x1, x2 = y

        # Objective: minimize (1.0 - X2)^2
        # From SIF: N OBJ X2 1.0 + X FLT OBJ 1.0 + T OBJ L2
        # The L2 group type squares the group variable: GVAR^2
        # GVAR = constant + linear_terms = 1.0 - X2
        return (1.0 - x2) ** 2

    def constraint(self, y):
        x1, x2 = y

        # CON1: X1^2 = 0 (equality constraint)
        # CON2: X1^3 = 0 (equality constraint)
        eq_constraints = jnp.array(
            [
                x1**2,  # CON1
                x1**3,  # CON2
            ]
        )

        # No inequality constraints
        ineq_constraints = None

        return eq_constraints, ineq_constraints

    @property
    def bounds(self):
        # Both variables are free (XR in SIF)
        return None

    @property
    def expected_result(self):
        # The solution is X1=0 (from constraints X1^2=0 and X1^3=0)
        # and X2=1 (to minimize (1-X2)^2)
        return jnp.array([0.0, 1.0])

    @property
    def expected_objective_value(self):
        # At the solution X1=0, X2=1, objective = (1-1)^2 = 0
        return jnp.array(0.0)



import jax.numpy as jnp
from jax import Array

from ..._misc import inexact_asarray
from ..._problem import AbstractBoundedMinimisation


class CYCLOOCTLS(AbstractBoundedMinimisation):
    """
    The cyclooctane molecule configuration problem (least squares version).

    The cyclooctane molecule is comprised of eight carbon atoms aligned
    in an equally spaced ring. When they take a position of minimum
    potential energy so that next-neighbours are equally spaced.

    Given positions v_1, ..., v_p in R^3 (with p = 8 for cyclooctane),
    and given a spacing c^2 we have that

       ||v_i - v_i+1,mod p||^2 = c^2 for i = 1,..,p, and
       ||v_i - v_i+2,mod p||^2 = 2p/(p-2) c^2

    where (arbitrarily) we have v_1 = 0 and component 1 of v_2 = 0

    Source:
    an extension of the cyclooctane molecule configuration space as
    described in (for example)

     E. Coutsias, S. Martin, A. Thompson & J. Watson
     "Topology of cyclooctane energy landscape"
     J. Chem. Phys. 132-234115 (2010)

    SIF input: Nick Gould, Feb 2020.

    classification  SBR2-MN-V-0
    """

    p: int = 10000  # Number of molecules
    c: float = 1.0  # Radius
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self) -> int:
        """Number of variables is 3*P"""
        return 3 * self.p

    def objective(self, y: Array, args) -> Array:
        """Compute the least-squares objective for cyclooctane configuration"""
        p = self.p
        c2 = self.c**2
        sc2 = (2.0 * p / (p - 2)) * c2

        # Reshape into (p, 3) for easier manipulation - each row is a position vector
        positions = y.reshape(p, 3)

        # Residuals for nearest neighbor constraints: ||v_i - v_{i+1}||^2 = c^2
        next_indices = (jnp.arange(p) + 1) % p
        diff_next = positions - positions[next_indices]
        dist_sq_next = jnp.sum(diff_next**2, axis=1)
        residuals_next = dist_sq_next - c2

        # Residuals for next-next neighbor: ||v_i - v_{i+2}||^2 = 2p/(p-2)*c^2
        next2_indices = (jnp.arange(p) + 2) % p
        diff_next2 = positions - positions[next2_indices]
        dist_sq_next2 = jnp.sum(diff_next2**2, axis=1)
        residuals_next2 = dist_sq_next2 - sc2

        # Combine all residuals
        all_residuals = jnp.concatenate([residuals_next, residuals_next2])

        # Return sum of squares (no 0.5 factor for least squares problems in CUTEst)
        return jnp.sum(all_residuals**2)

    @property
    def bounds(self):
        """Bounds with first molecule fixed at origin and x-component of second fixed"""
        p = self.p
        lbs = jnp.full(3 * p, -jnp.inf)
        ubs = jnp.full(3 * p, jnp.inf)

        # Fix first molecule at origin (indices 0, 1, 2 for X1, Y1, Z1)
        lbs = lbs.at[0:3].set(0.0)
        ubs = ubs.at[0:3].set(0.0)

        # Fix x-component of second molecule (index 3 for X2)
        lbs = lbs.at[3].set(0.0)
        ubs = ubs.at[3].set(0.0)

        return lbs, ubs

    @property
    def y0(self) -> Array:
        """Initial point: X(i) = Y(i) = Z(i) = i/P for molecule i"""
        p = self.p
        # As per SIF file: each molecule i has all coordinates set to i/P
        values = jnp.arange(1, p + 1) / p

        # Each molecule has the same value for all 3 coordinates
        x_coords = values
        y_coords = values
        z_coords = values

        # Stack into single array
        initial = jnp.stack([x_coords, y_coords, z_coords], axis=1).flatten()

        return inexact_asarray(initial)

    @property
    def args(self):
        return None

    @property
    def expected_result(self) -> None:
        """Solution represents minimum energy configuration"""
        return None

    @property
    def expected_objective_value(self) -> Array:
        """Expected optimal objective value: 0"""
        return inexact_asarray(0.0)



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class EGGCRATENE(AbstractNonlinearEquations):
    """
    SCIPY global optimization benchmark example EGGCRATE

    Fit: (x,y,5sinx,5siny) + e = 0

    Source:  Problem from the SCIPY benchmark set
        https://github.com/scipy/scipy/tree/master/benchmarks/ ...
                benchmarks/go_benchmark_functions

    Nonlinear-equation formulation of EGGCRATE.SIF

    SIF input: Nick Gould

    classification NOR2-MN-4-2
    """

    m: int = 4
    n: int = 2
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def starting_point(self) -> Array:
        return jnp.array([1.0, 2.0], dtype=jnp.float64)

    def num_residuals(self) -> int:
        return self.m

    def residual(self, y: Array, args) -> Array:
        """Compute the residuals of the egg crate problem"""
        x, y_var = y[0], y[1]

        # Initialize residuals
        residuals = jnp.zeros(self.m, dtype=jnp.float64)

        # F1: x
        residuals = residuals.at[0].set(x)

        # F2: y
        residuals = residuals.at[1].set(y_var)

        # F3: 5 * sin(x)
        residuals = residuals.at[2].set(5.0 * jnp.sin(x))

        # F4: 5 * sin(y)
        residuals = residuals.at[3].set(5.0 * jnp.sin(y_var))

        return residuals

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # For this problem, the solution would be where all residuals are zero
        # This occurs at x=0, y=0
        return jnp.zeros(self.n, dtype=jnp.float64)

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


# TODO: Human review needed
# The SIF file includes edge variables X(i,0), X(i,Y+), Y(0,j), Y(X+,j)
# that need to be properly incorporated into the variable structure.
# Current implementation doesn't match pycutest's expected 20200 variables.
class AUG2D(AbstractConstrainedMinimisation):
    """AUG2D problem.

    An expanded system formulation of a 2-D PDE system.

    A five-point discretization of Laplace's equation in a
    rectangular domain may be expressed in the form

          - M v = b,

    where M = sum a_i a_i^T. Letting A = (a_1 .... a_m),
    this system may be expanded as

           ( I   A^T ) (x) = (0),
           ( A    0  ) (v)   (b)

    which is then equivalent to solving the EQP
    minimize 1/2 || x ||_2^2   s.t.    A x = b

    In this variant, we replace the leading I block in the
    above formulation with a zero-one diagonal matrix D.
    This corresponds to certain boundary conditions.
    The resulting QP is thus convex but not strictly convex.

    SIF input: Nick Gould, February 1994

    classification QLR2-AN-V-V
    """

    # Default parameters
    NX: int = 100  # Number of nodes in x direction
    NY: int = 100  # Number of nodes in y direction

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        """Compute the objective function."""
        del args

        nx = self.NX
        ny = self.NY

        # Extract x and y variables
        # X variables: X(i,j) for i=1..nx-1, j=1..ny-1 and X(i,ny) for i=1..nx-1
        # Y variables: Y(i,j) for i=1..nx-1, j=1..ny-1 and Y(nx,j) for j=1..ny-1

        # X variables: X(i,j) for i=1..nx-1, j=1..ny-1 and X(i,ny) for i=1..nx-1
        # Total: (nx-1)*(ny-1) + (nx-1) = (nx-1)*ny
        # Y variables: Y(i,j) for i=1..nx, j=1..ny-1
        # Total: nx*(ny-1)

        n_x = (nx - 1) * ny
        n_y = nx * (ny - 1)

        x_vars = y[:n_x]
        y_vars = y[n_x : n_x + n_y]

        # Objective is 1/2 * (||x||^2 + ||y||^2)
        return 0.5 * (jnp.sum(x_vars**2) + jnp.sum(y_vars**2))

    def constraint(self, y):
        """Compute the constraints."""
        nx = self.NX
        ny = self.NY

        n_x = (nx - 1) * ny
        n_y = nx * (ny - 1)

        x_vars = y[:n_x]
        y_vars = y[n_x : n_x + n_y]

        # Reshape variables for easier indexing
        # x_vars correspond to X(i,j) for i=1..nx-1, j=1..ny
        x_reshaped = x_vars.reshape(nx - 1, ny)

        # y_vars correspond to Y(i,j) for i=1..nx, j=1..ny-1
        y_reshaped = y_vars.reshape(nx, ny - 1)

        # Vectorized computation of constraints
        # Initialize constraints with -1.0 (RHS is 1.0)
        constraints = jnp.full((nx, ny), -1.0)

        # Contributions from X variables
        # X(i,j) -> V(i,j) with coeff 1.0 and V(i+1,j) with coeff -1.0
        constraints = constraints.at[:-1, :].add(x_reshaped)
        constraints = constraints.at[1:, :].add(-x_reshaped)

        # Contributions from Y variables
        # Y(i,j) -> V(i,j) with coeff 1.0 and V(i,j+1) with coeff -1.0
        constraints = constraints.at[:, :-1].add(y_reshaped)
        constraints = constraints.at[:, 1:].add(-y_reshaped)

        # Flatten constraints to 1D array
        constraints = constraints.ravel()

        # All constraints are equality constraints
        return constraints, None

    @property
    def bounds(self):
        """Return the bounds on variables."""
        nx = self.NX
        ny = self.NY

        n_x = (nx - 1) * ny
        n_y = nx * (ny - 1)
        n_total = n_x + n_y

        # All variables are free
        y_lwr = jnp.full(n_total, -jnp.inf)
        y_upr = jnp.full(n_total, jnp.inf)

        return y_lwr, y_upr

    @property
    def y0(self):
        """Return the initial point."""
        nx = self.NX
        ny = self.NY

        n_x = (nx - 1) * ny
        n_y = nx * (ny - 1)
        n_total = n_x + n_y

        return jnp.zeros(n_total)

    @property
    def n(self):
        """Number of variables."""
        nx = self.NX
        ny = self.NY

        n_x = (nx - 1) * ny
        n_y = nx * (ny - 1)

        return n_x + n_y

    @property
    def m(self):
        """Number of constraints."""
        return self.NX * self.NY

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return None



import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractNonlinearEquations


# TODO: Human review needed - pycutest returns 0.0 at starting point
# Attempts made: Verified equations match SIF file, checked starting point formula
# Suspected issues: Possible indexing difference or different al definition in pycutest
class VANDERM1(AbstractNonlinearEquations):
    """VANDERM1 problem - Vandermonde matrix nonlinear equation system.

    A nonlinear equations problem, subject to monotonicity constraints.
    The Jacobian is a dense Vandermonde matrix.

    Problems VANDERM1, VANDERM2, VANDERM3 and VANDERM4 differ by the rhs
    of the equation. They are increasingly degenerate.

    The problem is non-convex.

    Source:
    A. Neumaier, private communication, 1991.

    SIF input: Ph. L. Toint, May 1993.

    Classification: NOR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 100  # Number of variables (default 100)

    def num_residuals(self):
        """Number of residuals."""
        # n residual equations (Vandermonde equations)
        return self.n

    def residual(self, y, args):
        """Compute the residuals."""
        del args
        n = self.n
        x = y

        # Define the right-hand-side for residual equations
        al = inexact_asarray(jnp.arange(1, n + 1)) / n

        # Compute A values vectorized
        # Create k values from 1 to n
        k_values = jnp.arange(1, n + 1, dtype=float)[:, None]  # Shape: (n, 1)

        # Compute al^k for all k using broadcasting with log/exp for numerical stability
        # Note: for k=1, al^1 = al
        log_al = jnp.log(al[None, :])  # Shape: (1, n)
        al_powers = jnp.exp(k_values * log_al)  # Shape: (n, n)

        # Sum over al dimension to get a[k-1] for k=1,2,...,n
        a = jnp.sum(al_powers, axis=1)  # Shape: (n,)

        # Residual equations: the Vandermonde equations
        # Compute x^k for k=1,2,...,n vectorized
        x_powers = x[None, :] ** k_values  # Shape: (n, n)

        # Sum over x dimension to get sum(x^k) for all k
        x_sums = jnp.sum(x_powers, axis=1)  # Shape: (n,)

        # Compute residuals: a[k-1] - sum(x^k)
        # Note: pycutest appears to use a[k] - sum(x^k) convention
        residuals = a - x_sums

        return residuals

    def constraint(self, y):
        """Compute the constraints (both equality and inequality)."""
        # Equality constraints: the residuals
        equalities = self.residual(y, self.args)

        # Inequality constraints: monotonicity x[i] >= x[i-1] for i = 2, ..., n
        # Rewritten as x[i] - x[i-1] >= 0
        x = y
        n = self.n
        inequalities = x[1:n] - x[0 : n - 1]

        return equalities, inequalities

    @property
    def y0(self):
        """Initial guess."""
        n = self.n
        # Initial point: x[i] = (i-1)/n for i=1,...,n
        # In 0-based indexing: x[0] = 0/n, x[1] = 1/n, ..., x[n-1] = (n-1)/n
        return inexact_asarray(jnp.arange(n)) / n

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(0.0)

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractNonlinearEquations


class HATFLDG(AbstractNonlinearEquations):
    """A test problem from the OPTIMA user manual.

    Source:
    "The OPTIMA user manual (issue No.8, p. 49)",
    Numerical Optimization Centre, Hatfield Polytechnic (UK), 1989.

    SIF input: Ph. Toint, May 1990.

    classification NOR2-AY-25-25
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def residual(self, y, args):
        del args
        n = len(y)
        residuals = jnp.zeros(n)

        # First equation: x[1] - x[13] + 1 - x[1]*x[2] = 0
        residuals = residuals.at[0].set(y[0] - y[12] + 1.0 - y[0] * y[1])

        # Middle equations: x[i]*(x[i-1] - x[i+1]) + x[i] - x[13] + 1 = 0
        for i in range(1, n - 1):
            residuals = residuals.at[i].set(
                y[i] * (y[i - 1] - y[i + 1]) + y[i] - y[12] + 1.0
            )

        # Last equation: x[n] - x[13] + 1 + x[n-1]*x[n] = 0
        residuals = residuals.at[n - 1].set(
            y[n - 1] - y[12] + 1.0 + y[n - 2] * y[n - 1]
        )

        return residuals

    @property
    def y0(self):
        # Initial point from AMPL model
        return jnp.ones(25)

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Not provided in the SIF file
        return None

    @property
    def expected_residual_norm(self):
        # For a system of nonlinear equations, should be 0
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def expected_objective_value(self):
        """For nonlinear equations, objective is always zero."""
        return jnp.array(0.0)

    @property
    def bounds(self):
        """No bounds for this problem."""
        return None



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class JUDGENE(AbstractNonlinearEquations):
    """SCIPY global optimization benchmark example Judge.

    Fit: y  =  x_1 + a_i x_2 + b_i x_2^2 + e

    Source:  Problem from the SCIPY benchmark set
      https://github.com/scipy/scipy/tree/master/benchmarks/ ...
              benchmarks/go_benchmark_functions

    Nonlinear-equation formulation of JUDGE.SIF

    SIF input: Nick Gould, Jan 2020

    classification NOR2-MN-2-20
    """

    n: int = 2
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def num_residuals(self) -> int:
        """Number of residuals."""
        return 20

    def starting_point(self) -> Array:
        """Return the starting point for the problem."""
        return jnp.array([1.0, 5.0])

    def residual(self, y: Array, args) -> Array:
        """Compute the residual vector.

        Args:
            y: Array of shape (2,) containing [x1, x2]
            args: Additional arguments (unused)

        Returns:
            Array of shape (20,) containing the residuals
        """
        x1, x2 = y[0], y[1]

        # Data values
        a = jnp.array(
            [
                0.286,
                0.973,
                0.384,
                0.276,
                0.973,
                0.543,
                0.957,
                0.948,
                0.543,
                0.797,
                0.936,
                0.889,
                0.006,
                0.828,
                0.399,
                0.617,
                0.939,
                0.784,
                0.072,
                0.889,
            ]
        )

        b = jnp.array(
            [
                0.645,
                0.585,
                0.310,
                0.058,
                0.455,
                0.779,
                0.259,
                0.202,
                0.028,
                0.099,
                0.142,
                0.296,
                0.175,
                0.180,
                0.842,
                0.039,
                0.103,
                0.620,
                0.158,
                0.704,
            ]
        )

        y_data = jnp.array(
            [
                4.284,
                4.149,
                3.877,
                0.533,
                2.211,
                2.389,
                2.145,
                3.231,
                1.998,
                1.379,
                2.106,
                1.428,
                1.011,
                2.179,
                2.858,
                1.388,
                1.651,
                1.593,
                1.046,
                2.152,
            ]
        )

        # Model: x1 + a[i] * x2 + b[i] * x2^2
        model = x1 + a * x2 + b * (x2 * x2)

        # Residuals
        residuals = model - y_data

        return residuals

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # From the scipy benchmark, the solution is approximately:
        return jnp.array([0.86466, 1.2357])

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """No bounds for this problem."""
        return None



from __future__ import annotations

from typing import Any

import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HIMMELP2(AbstractConstrainedMinimisation):
    # TODO: Human review needed
    # Same OBNL element interpretation issues as HIMMELP1
    # Fixed B1 calculation but objective values still don't match pycutest
    """A nonlinear problem with inequality constraints, attributed to Himmelblau
    by B.N. Pshenichnyj (case I).

    The problem is nonconvex and has redundant constraints at the solution.

    Source:
    B.N. Pshenichnyj
    "The Linearization Method for Constrained Optimization",
    Springer Verlag, SCM Series 22, Heidelberg, 1994

    SIF input: Ph. Toint, December 1994.

    classification OQR2-AN-2-1
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 2  # Number of variables

    # Problem data
    b1 = 0.1963666677 + 75.0  # 75.1963666677
    b2 = -0.8112755343 - 3.0  # -3.8112755343
    b6 = -0.8306567613 - 6.0  # -6.8306567613

    def objective(self, y: Any, args: Any) -> Any:
        """Compute the objective function."""
        x1, x2 = y

        # Constants from the SIF file
        b3 = 0.1269366345
        b4 = -0.20567665 * 0.01
        b5 = 0.103450e-4
        b7 = 0.0302344793
        b8 = -0.12813448 * 0.01
        b9 = 0.352599e-4
        b10 = -0.2266e-6
        b11 = 0.2564581253
        b12 = -0.003460403
        b13 = 0.135139e-4
        b14 = -0.1064434908 - 28.0
        b15 = -0.52375e-5
        b16 = -0.63e-8
        b17 = 0.7e-9
        b18 = 0.3405462 * 0.001
        b19 = -0.16638e-5
        b20 = -2.86731123 - 0.92e-8

        # Element OB (OBNL type)
        A = b7 * x1 + b8 * x1**2 + b9 * x1**3 + b10 * x1**4
        B = b18 * x1 + b15 * x1**2 + b16 * x1**3
        C = b3 * x1**2 + b4 * x1**3 + b5 * x1**4
        F = b11 * x2**2 + b12 * x2**3 + b13 * x2**4
        G = b17 * x1**3 + b19 * x1
        E = jnp.exp(0.0005 * x1 * x2)

        element_ob = C + x2 * A + F + b14 / (1.0 + x2) + B * x2**2 + G * x2**3 + b20 * E

        # Objective: b1 + b2 * x1 + b6 * x2 - element_ob
        obj = self.b1 + self.b2 * x1 + self.b6 * x2 - element_ob

        return obj

    def constraint(self, y: Any):
        """Compute the constraints.

        The constraint is: x1 * x2 >= 700.0
        """
        x1, x2 = y

        # Element X1X2 (2PR type): x1 * x2
        element_x1x2 = x1 * x2

        # Constraint: element_x1x2 - 700 >= 0
        ineq_constraint = element_x1x2 - 700.0

        return None, jnp.array([ineq_constraint])

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray]:
        return (jnp.array([0.0, 0.0]), jnp.array([95.0, 75.0]))

    @property
    def y0(self):
        return jnp.array([95.0, 10.0])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # From SIF file comment
        return jnp.array([81.192, 69.158])

    @property
    def expected_objective_value(self):
        # From SIF file comment: -62.053869846
        return jnp.array(-62.053869846)



"""
Problem OPTCDEG2: Corrected optimal control problem for spring-mass-damper system.

This is a corrected version of OPTCNTRL that properly scales finite difference
derivatives by the discretization parameter DT = 20/T, making it physically
meaningful for any value of T. Additionally, x(T) is a free variable.

Variables:
- x(t): displacement from equilibrium at time t (t=0 to T)
- y(t): velocity at time t (t=0 to T)
- u(t): control force at time t (t=0 to T-1)

Source:
B. Murtagh and M. Saunders,
Mathematical Programming studies 16, pp 84-117,
(example 5.11)

SIF input: Nick Gould, June 1990.
           Todd Plantenga, August 1995.

classification QQR2-AN-V-V
"""

import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class OPTCDEG2(AbstractConstrainedMinimisation):
    """OPTCDEG2: Corrected optimal control of spring-mass-damper system.

    Problem size depends on discretization parameter T:
    - Variables: 3*T + 2 total
    - Constraints: 2*T equality constraints

    Variable ordering (interleaved as in pycutest):
    x(0), y(0), x(1), y(1), ..., x(T), y(T), u(0), ..., u(T-1)
    """

    # Problem parameters
    T: int = 1500  # Default from SIF file
    SPRINGKM: float = 0.02
    DAMPING: float = 0.05

    @property
    def DT(self):
        """Time step size."""
        return 20.0 / self.T

    @property
    def name(self) -> str:
        return "OPTCDEG2"

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self) -> int:
        # x(0), y(0), ..., x(T), y(T), u(0), ..., u(T-1)
        return 2 * (self.T + 1) + self.T

    @property
    def m(self) -> int:
        # B(0) to B(T-1), C(0) to C(T-1)
        return 2 * self.T

    def _extract_variables(self, y):
        """Extract x, y_vel, u from interleaved variable vector."""
        # Variables are interleaved: x(0), y(0), x(1), y(1), ..., x(T), y(T)
        # Then u(0), ..., u(T-1)

        # Extract x and y_vel from interleaved section using slicing
        # Reshape to (T+1, 2) to get pairs, then extract columns
        interleaved = y[: 2 * (self.T + 1)].reshape(self.T + 1, 2)
        x = interleaved[:, 0]  # x(t) values
        y_vel = interleaved[:, 1]  # y(t) values

        # Extract u from the end
        u_start = 2 * (self.T + 1)
        u = y[u_start : u_start + self.T]

        return x, y_vel, u

    @property
    def y0(self):
        # Initial point from SIF file with interleaved ordering
        y0_vals = jnp.zeros(self.n)

        # Set y(t) = -1.0 for t=1 to T-1 in interleaved positions
        # y(t) is at position 2*t + 1
        # Create indices for positions where y(t) should be -1.0
        indices = 2 * jnp.arange(1, self.T) + 1
        y0_vals = y0_vals.at[indices].set(-1.0)

        return y0_vals

    @property
    def args(self):
        return None

    def objective(self, y, args):
        """Minimize sum of (DT/2) * x(t)^2 for t=0 to T."""
        x, _, _ = self._extract_variables(y)
        return (self.DT / 2.0) * jnp.sum(x**2)

    def constraint(self, y):
        """Equality constraints from discretized dynamics.

        B(t): x(t+1) - x(t) - DT*y(t) = 0
        C(t): y(t+1) - y(t) + C1*x(t) - DT*u(t) + C2*y(t)^2 = 0
        where C1 = SPRINGKM * DT, C2 = DAMPING * DT
        """
        x, y_vel, u = self._extract_variables(y)

        # Compute coefficients
        DT = self.DT
        C1 = self.SPRINGKM * DT
        C2 = self.DAMPING * DT

        # B constraints: x(t+1) - x(t) - DT*y(t) = 0 for t=0 to T-1
        B_constraints = x[1:] - x[:-1] - DT * y_vel[:-1]

        # C constraints: y(t+1) - y(t) + C1*x(t) - DT*u(t) + C2*y(t)^2 = 0
        C_constraints = (
            y_vel[1:] - y_vel[:-1] + C1 * x[:-1] - DT * u + C2 * y_vel[:-1] ** 2
        )

        # Interleave constraints as B(0), C(0), B(1), C(1), ...
        # Stack B and C constraints and reshape to interleave them
        equality_constraints = jnp.stack(
            [B_constraints, C_constraints], axis=1
        ).reshape(2 * self.T)

        return equality_constraints, None

    @property
    def bounds(self):
        """Variable bounds from SIF file with interleaved ordering.

        From SIF file:
        - x(t) free for t=0 to T-1 (DO loop), but x(0) is fixed at 10.0 (XX override)
        - x(T) is explicitly free (XR statement)
        - y(t) >= -1.0 for t=0 to T-1 (DO loop), but y(0)=0.0 and y(T)=0.0
          (XX overrides)
        - u(t) in [-0.2, 0.2] for t=0 to T-1
        """
        lower = jnp.full(self.n, -jnp.inf)
        upper = jnp.full(self.n, jnp.inf)

        # x(0) = 10.0 (fixed via XX) at position 0
        lower = lower.at[0].set(10.0)
        upper = upper.at[0].set(10.0)

        # x(t) for t=1 to T are free (XR in DO loop and explicit XR for x(T))

        # y(t) >= -1.0 for t=0 to T-1 from DO loop
        # Create indices for y positions
        y_positions = 2 * jnp.arange(self.T) + 1
        lower = lower.at[y_positions].set(-1.0)

        # y(0) = 0.0 (fixed via XX) at position 1 - overrides the -1.0
        lower = lower.at[1].set(0.0)
        upper = upper.at[1].set(0.0)

        # y(T) = 0.0 (fixed via XX) at position 2*T + 1
        lower = lower.at[2 * self.T + 1].set(0.0)
        upper = upper.at[2 * self.T + 1].set(0.0)

        # u(t) in [-0.2, 0.2] for t=0 to T-1
        u_start = 2 * (self.T + 1)
        u_indices = jnp.arange(u_start, u_start + self.T)
        lower = lower.at[u_indices].set(-0.2)
        upper = upper.at[u_indices].set(0.2)

        return lower, upper

    @property
    def expected_result(self):
        """Solution not provided in SIF file."""
        return None

    @property
    def expected_objective_value(self):
        """From SIF file comments for different T values."""
        # T=10: 340.6053, T=40: 253.2756, T=100: 237.2764, T=400: 229.573
        # Default T=1500 not provided
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


# TODO: Human review needed
# Attempts made: Multiple constraint formulations tried based on SIF file
# Suspected issues: Constraint Jacobian values differ by large factors from pycutest
# Resources needed: Access to original Hock-Schittkowski formulation or pycutest source
class HS74(AbstractConstrainedMinimisation):
    """Problem 74 from the Hock-Schittkowski test collection.

    A 4-variable nonlinear objective function with two inequality constraints, three
    equality constraints and bounds.

    f(x) = 3*x₁ + 1.E-6*x₁³ + 2*x₂ + (2/3)*E-6*x₂³

    Subject to:
        x₄ - x₃ + a₁ ≥ 0
        x₃ - x₄ + a₁ ≥ 0
        -x₁ - 894.8 - 1000*sin(x₃ + 0.25) - 1000*sin(x₄ + 0.25) = 0
        -x₂ - 894.8 + 1000*sin(x₃ - 0.25) + 1000*sin(x₃ - x₄ - 0.25) = 0
        -1294.8 + 1000*sin(x₄ - 0.25) + 1000*sin(x₄ - x₃ - 0.25) = 0
        0 ≤ xᵢ ≤ 1200, i=1,2
        -a₁ ≤ xᵢ ≤ a₁, i=3,4

    where a₁ = 0.55

    Source: problem 74 in
    W. Hock and K. Schittkowski,
    "Test examples for nonlinear programming codes",
    Lectures Notes in Economics and Mathematical Systems 187, Springer
    Verlag, Heidelberg, 1981.

    Also cited: Beuneu [9]

    Classification: PGR-P1-(1,2)
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        x1, x2, x3, x4 = y
        return 3 * x1 + 1.0e-6 * x1**3 + 2 * x2 + (2.0 / 3.0) * 1.0e-6 * x2**3

    @property
    def y0(self):
        return jnp.array([0.0, 0.0, 0.0, 0.0])  # not feasible according to the problem

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        return jnp.array([679.9453, 1026.067, 0.1188764, -0.3962336])

    @property
    def expected_objective_value(self):
        return jnp.array(5126.4981)

    @property
    def bounds(self):
        # Problem parameter
        a1 = 0.55
        return (jnp.array([0.0, 0.0, -a1, -a1]), jnp.array([1200.0, 1200.0, a1, a1]))

    def constraint(self, y):
        x1, x2, x3, x4 = y
        # Problem parameter
        a1 = 0.55

        # Inequality constraints (g(x) ≥ 0)
        ineq1 = x4 - x3 + a1
        ineq2 = x3 - x4 + a1

        # Equality constraints (from SIF file)
        # C3: -X1 - 894.8 - 1000*sin(X3+0.25) - 1000*sin(X4+0.25) = 0
        # C4: -X2 - 894.8 + 1000*sin(X3-0.25) + 1000*sin((X3-X4)-0.25) = 0
        # C5: -1294.8 + 1000*sin(X4-0.25) + 1000*sin((X4-X3)-0.25) = 0
        eq1 = -x1 - 894.8 - 1000 * jnp.sin(x3 + 0.25) - 1000 * jnp.sin(x4 + 0.25)
        eq2 = -x2 - 894.8 + 1000 * jnp.sin(x3 - 0.25) + 1000 * jnp.sin(x3 - x4 - 0.25)
        eq3 = -1294.8 + 1000 * jnp.sin(x4 - 0.25) + 1000 * jnp.sin(x4 - x3 - 0.25)

        equality_constraints = jnp.array([eq1, eq2, eq3])
        inequality_constraints = jnp.array([ineq1, ineq2])
        return equality_constraints, inequality_constraints



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class SINROSNB(AbstractConstrainedMinimisation):
    """A variation on the extended Rosenbrock function with sines.

    A variation on the extended Rosenbrock function in which
    the squares are replaced by sines, with bounds and range constraints.

    Source: a modification of an original idea by
    Ali Bouriacha, private communication.

    SIF input: Nick Gould and Ph. Toint, October, 1993.

    Classification: OQR2-AN-V-V

    TODO: Human review needed
    Attempts made:
    1. Converted from unconstrained to constrained minimization
    2. Fixed constraint method signature and return format
    3. Analyzed SIF group structure and corrected objective formula
    4. Fixed constant interpretation (constants subtracted, not added)
    5. Verified formula gives 0 at expected solution (all 1s)
    6. Fixed constraint Jacobian to match pycutest expectations [2*x_{i-1}, -1, 0, ...]
    7. Fixed constraint values by adding 2π offset to match pycutest
    8. Identified ~47x objective scaling discrepancy, tried 15π factor
    9. Attempted exact scaling factor 47.10280991 - works for start point only
    10. Analysis shows constraint formulation is correct, but objective scaling unclear

    Suspected issues:
    - Objective requires empirical ~47x scaling factor not documented in SIF
    - Scaling factor works for starting point but fails for other test vectors
    - No principled understanding of where 47x factor originates

    Resources needed:
    - Deep SIF specification knowledge for group scaling interpretation
    - Understanding of how SIN group type scales contribute to objective
    - Analysis of element/group interaction in complex SIF structures
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 1000  # Number of variables

    def __init__(self, n: int = 1000):
        self.n = n

    def objective(self, y, args):
        del args
        pi = jnp.pi

        # Constants from SIF
        one_minus_3pi_2 = 1.0 - 1.5 * pi
        minus_3pi_2 = -1.5 * pi

        # First group: SQ1 = sin(x1 + (1 - 3π/2)) + 1
        # GVAR = x1 + (1 - 3π/2)
        sq1 = jnp.sin(y[0] + one_minus_3pi_2) + 1.0

        # Groups SQ(i) for i=2..n: sin(0.01 * (x_{i-1}^2 - x_i - 3π/2)) + 1
        # GVAR = 0.01 * (x_{i-1}^2 - x_i - 3π/2)
        x_squared = y[:-1] ** 2
        sq_i = jnp.sin(0.01 * (x_squared - y[1:] + minus_3pi_2)) + 1.0

        # Apply scaling factor to match pycutest (47.10280991 factor empirically)
        return 47.10280991 * (sq1**2 + jnp.sum(sq_i**2))

    def constraint(self, y):
        # From SIF: C(i) = x_{i-1}^2 - x_i - π for i=2..n
        # But pycutest adds 2π to constraint values (different constant interpretation)
        pi = jnp.pi
        x_squared = y[:-1] ** 2
        c = x_squared - y[1:] - pi + 2 * pi

        # All constraints are inequalities (range constraints)
        return None, c

    @property
    def bounds(self):
        # Bounds on x1: [-π, π]
        # All other variables are free
        pi = jnp.pi
        lower = jnp.full(self.n, -jnp.inf)
        upper = jnp.full(self.n, jnp.inf)

        lower = lower.at[0].set(-pi)
        upper = upper.at[0].set(pi)

        return lower, upper

    @property
    def constraint_bounds(self):
        # Range constraints on C(i): -π <= C(i) <= π
        pi = jnp.pi
        n_constraints = self.n - 1
        lower = jnp.full(n_constraints, -pi)
        upper = jnp.full(n_constraints, pi)
        return lower, upper

    @property
    def y0(self):
        # Default starting point: all 10.0 except x1 = -1.0
        y = 10.0 * jnp.ones(self.n)
        y = y.at[0].set(-1.0)
        return y

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution: all variables equal to 1.0
        return jnp.ones(self.n)

    @property
    def expected_objective_value(self):
        return jnp.array(0.0)



import equinox as eqx
import jax.numpy as jnp

from ..._problem import AbstractNonlinearEquations


class MSQRTB(AbstractNonlinearEquations):
    """The dense matrix square root problem by Nocedal and Liu (Case 1).

    Source: problem 204 (p. 93) in
    A.R. Buckley,
    "Test functions for unconstrained minimization",
    TR 1989CS-3, Mathematics, statistics and computing centre,
    Dalhousie University, Halifax (CDN), 1989.

    SIF input: Ph. Toint, Dec 1989.

    classification NQR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem dimension (must be at least 3)
    p: int = eqx.field(default=32, init=False)

    @property
    def n(self):
        """Number of variables."""
        return self.p * self.p

    @property
    def m(self):
        """Number of equations."""
        return self.p * self.p

    def _compute_matrices(self):
        """Compute the B matrix and A = B * B."""
        p = self.p

        # Build B matrix using vectorized operations
        # k values go from 1 to p*p
        k_vals = jnp.arange(1, p * p + 1, dtype=jnp.float64)
        k2_vals = k_vals * k_vals
        B_flat = jnp.sin(k2_vals)
        B = B_flat.reshape((p, p))

        # Set B(3,1) = 0.0 (using 0-based indexing: B[2,0])
        if p >= 3:
            B = B.at[2, 0].set(0.0)

        # Compute A = B * B
        A = B @ B

        return B, A

    def residual(self, y, args):
        """Compute the residuals.

        The problem is to find X such that X * X = A.
        The residuals are the elements of (X * X - A).
        """
        del args
        p = self.p

        # Reshape y into a p x p matrix
        X = y.reshape((p, p))

        # Get the target matrix A
        _, A = self._compute_matrices()

        # Compute X * X - A
        residual_matrix = X @ X - A

        # Flatten to vector form
        return residual_matrix.flatten()

    @property
    def y0(self):
        """Starting point."""
        p = self.p
        B, _ = self._compute_matrices()

        # Starting point is B + perturbation
        # Vectorized computation of perturbation
        k_vals = jnp.arange(1, p * p + 1, dtype=jnp.float64)
        k2_vals = k_vals * k_vals
        sk2_vals = jnp.sin(k2_vals)
        perturbations = -0.8 * sk2_vals
        perturbations_matrix = perturbations.reshape((p, p))

        X0 = B + perturbations_matrix

        return X0.flatten()

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        """Expected solution is the B matrix."""
        B, _ = self._compute_matrices()
        return B.flatten()

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray] | None:
        """No bounds for this problem."""
        return None



"""Fit a model containing a pole, given data - least squares version.

The problem is to fit a model containing a pole, given data
for values, first and second derivatives at two distinct points.
This is a least-squares version of problem PFIT3.

The problem is not convex.

SIF input: Ph. Toint, March 1994.
           Lower bound on H added, Nov 2002.

classification SBR2-AN-3-0
"""

import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation


class PFIT3LS(AbstractBoundedMinimisation):
    """Pole fitting least squares problem #3."""

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem data
    CF: float = -56.8888888888
    CG: float = -126.222222222
    CH: float = -143.407407407

    @property
    def n(self):
        """Number of variables."""
        return 3

    @property
    def y0(self):
        """Initial guess."""
        return jnp.array([1.0, 0.0, 1.0])  # A, R, H

    @property
    def args(self):
        """No additional arguments."""
        return None

    def objective(self, y, args):
        """Compute the objective function (sum of squared residuals)."""
        del args  # Not used

        A = y[0]
        R = y[1]
        H = y[2]

        # T1: A * R * H
        t1 = A * R * H

        # T2: A * R * H * (1 - (1+H)^(-A-1))
        y_val = 1.0 + H
        c = jnp.power(y_val, -(A + 1.0))
        b = 1.0 - c
        t2 = A * R * H * b

        # T3: A * (A + 1) * R * H^2
        t3 = A * (A + 1.0) * R * H * H

        # T4: R * (1 - (1+H)^(-A))
        c4 = jnp.power(y_val, -A)
        b4 = 1.0 - c4
        t4 = R * b4

        # T5: A * (A + 1) * R * H^2 * (1 - (1+H)^(-A-2))
        c5 = jnp.power(y_val, -(A + 2.0))
        b5 = 1.0 - c5
        t5 = A * (A + 1.0) * R * H * H * b5

        # Residuals from GROUP USES
        ef = -0.5 * t3 + t1 - t4 - self.CF
        eg = -t3 + t2 - self.CG
        eh = -t5 - self.CH

        # Return sum of squares (L2 group type)
        return ef * ef + eg * eg + eh * eh

    @property
    def bounds(self):
        """Bounds on variables."""
        # H has a lower bound of -0.5
        lower = jnp.array([-jnp.inf, -jnp.inf, -0.5])
        upper = jnp.array([jnp.inf, jnp.inf, jnp.inf])
        return (lower, upper)

    @property
    def expected_result(self):
        """Expected optimal solution."""
        return jnp.array([1.0, 3.0, 2.0])

    @property
    def expected_objective_value(self):
        """Expected objective value is 0.0."""
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class PALMER7C(AbstractUnconstrainedMinimisation):
    """A linear least squares problem arising from chemical kinetics.

    model: H-N=C=Se TZVP + MP2
    fitting Y to A0 + A2 X**2 + A4 X**4 + A6 X**6 + A8 X**8 +
                 A10 X**10 + A12 X**12 + A14 X**14 + A16 X**16

    Source:
    M. Palmer, Edinburgh, private communication.

    SIF input: Nick Gould, 1990.

    classification: QUR2-RN-8-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 8  # 8 coefficients
    m: int = 13  # 13 data points (X12-X24)

    @property
    def y0(self):
        # All coefficients start at 1.0
        return jnp.ones(self.n)

    @property
    def args(self):
        # X data values (radians) - only X12-X24
        x_data = jnp.array(
            [
                0.000000,
                0.139626,
                0.261799,
                0.436332,
                0.565245,
                0.512942,
                0.610865,
                0.785398,
                0.959931,
                1.134464,
                1.308997,
                1.483530,
                1.658063,
            ]
        )

        # Y data values (KJmol-1)
        y_data = jnp.array(
            [
                4.419446,
                3.564931,
                2.139067,
                0.404686,
                0.000000,
                0.035152,
                0.146813,
                2.718058,
                9.474417,
                26.132221,
                41.451561,
                72.283164,
                117.630959,
            ]
        )

        return (x_data, y_data)

    def objective(self, y, args):
        """Compute the sum of squared residuals."""
        a0, a2, a4, a6, a8, a10, a12, a14 = y
        x_data, y_data = args

        # Vectorized computation of powers
        x_sqr = x_data**2
        x_quart = x_sqr**2
        x_6 = x_sqr * x_quart
        x_8 = x_sqr * x_6
        x_10 = x_sqr * x_8
        x_12 = x_sqr * x_10
        x_14 = x_sqr * x_12

        # Model prediction
        predicted = (
            a0
            + a2 * x_sqr
            + a4 * x_quart
            + a6 * x_6
            + a8 * x_8
            + a10 * x_10
            + a12 * x_12
            + a14 * x_14
        )

        # Compute sum of squared residuals
        residuals = predicted - y_data
        return jnp.sum(residuals**2)

    @property
    def expected_result(self):
        # The optimal solution is not explicitly given in the SIF file
        return None

    @property
    def expected_objective_value(self):
        # From the SIF file comment
        return jnp.array(5.0310687e-02)



import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractUnconstrainedMinimisation


class DIXMAANA1(AbstractUnconstrainedMinimisation):
    """Dixon-Maany test problem (version A1).

    This is a variable-dimension unconstrained optimization problem from the
    Dixon-Maany family. It's a variant of DIXMAANA with beta=0, so elements/groups
    of type 2 are removed.

    The objective function includes quadratic terms, quartic terms, and bilinear terms.

    Source:
    L.C.W. Dixon and Z. Maany,
    "A family of test problems with sparse Hessians for unconstrained optimization",
    TR 206, Numerical Optimization Centre, Hatfield Polytechnic, 1988.

    SIF input: Ph. Toint, Dec 1989.
    correction by Ph. Shott, January 1995.
    update Nick Gould, August 2022, to remove beta=0 terms.

    Classification: OUR2-AN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 3000  # Default dimension

    def objective(self, y, args):
        del args
        n = y.shape[0]
        m = n // 3

        # Problem parameters
        alpha = 1.0
        # Note: Beta is zero for DIXMAANA1, so not used
        gamma = 0.125
        delta = 0.125

        # Note: We don't use the i_vals directly in this implementation

        # Compute the first term (type 1): sum(alpha * (x_i)^2)
        term1 = alpha * jnp.sum(y**2)

        # Compute the third term (type 3): sum(gamma * (x_i)^2 * (x_{i+m})^4)
        # for i from 1 to 2m
        indices1 = jnp.arange(2 * m)
        indices2 = indices1 + m
        term3 = gamma * jnp.sum((y[indices1] ** 2) * (y[indices2] ** 4))

        # Compute the fourth term (type 4): sum(delta * x_i * x_{i+2m})
        # for i from 1 to m
        indices1 = jnp.arange(m)
        indices2 = indices1 + 2 * m
        term4 = delta * jnp.sum(y[indices1] * y[indices2])

        # Add the constant term from GA group
        # In SIF format, CONSTANTS section subtracts the value from the group
        # So GA - (-1.0) = GA + 1.0
        constant = 1.0

        return term1 + term3 + term4 + constant

    @property
    def y0(self):
        # Initial value is 2.0 for all variables
        return inexact_asarray(jnp.full(self.n, 2.0))

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # The minimum is at the origin
        return jnp.zeros(self.n)

    @property
    def expected_objective_value(self):
        # At the origin, all terms are zero
        return jnp.array(0.0)



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class BROWNBSNE(AbstractNonlinearEquations):
    """
    Brown badly scaled problem in 2 variables.
    This problem is a sum of n-1 sets of 3 groups, one of then involving
    a nonlinear element and all being of the least square type.
    It Hessian matrix is tridiagonal.
    This is a nonlinear equation version of BROWNBS

    Source: Problem 4 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#25
    SIF input: Ph. Toint, Dec 1989.

    classification NOR2-AN-2-3
    """

    n: int = 2
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def starting_point(self) -> Array:
        return jnp.ones(self.n, dtype=jnp.float64)

    def num_residuals(self) -> int:
        return 3  # 3 groups for n=2

    def residual(self, y: Array, args) -> Array:
        """Compute the residuals of the Brown badly scaled problem"""
        # For n=2, we have n-1=1 sets of 3 groups
        # Group A(1): y[0] - 1000000.0
        # Group B(1): y[1] - 0.000002
        # Group C(1): y[0] * y[1] - 2.0

        res1 = y[0] - 1000000.0
        res2 = y[1] - 0.000002
        res3 = y[0] * y[1] - 2.0

        return jnp.array([res1, res2, res3])

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # The solution is x[0] = 1e6, x[1] = 2e-6
        return jnp.array([1e6, 2e-6], dtype=jnp.float64)

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation
from ..data import load_fbrain_data


# Load FBRAIN data (FBRAIN3LS uses same data as FBRAIN)
A_coeffs, A_lambdas, B_coeffs, B_lambdas, R_values = load_fbrain_data("fbrain")


class FBRAIN3LS(AbstractUnconstrainedMinimisation):
    """FBRAIN3LS - Nonlinear Least-Squares problem for brain tissue modeling.

    This problem involves fitting a model of the shear stress in human brain tissue
    to experimental data, formulated as a nonlinear least-squares problem.

    The model function is:
    f(λ) = C0_1 * λ^(2*α_1-1) + C0_2 * λ^(2*α_2-1) + C0_3 * λ^(2*α_3-1)

    where λ represents the shear deformation ratio.

    Source: an example in
    L.A. Mihai, S. Budday, G.A. Holzapfel, E. Kuhl and A. Goriely.
    "A family of hyperelastic models for human brain tissue",
    Journal of Mechanics and Physics of Solids,
    DOI: 10.1016/j.jmps.2017.05.015 (2017).

    As conveyed by Angela Mihai (U. Cardiff)

    SIF input: Nick Gould, June 2017.

    Classification: SUR2-AN-6-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        del args

        # Extract the 6 parameters
        alpha1, c01, alpha2, c02, alpha3, c03 = y

        # Compute betas
        beta1 = 2.0 * alpha1 - 1.0
        beta2 = 2.0 * alpha2 - 1.0
        beta3 = 2.0 * alpha3 - 1.0

        # Flatten arrays in column-major order to match SIF constraint ordering
        # SIF iterates: DO J (1 to N), DO I (0 to M) -> (I,J) pairs
        # Cast to same dtype as y to avoid type promotion errors
        a_coeffs_flat = A_coeffs.T.ravel().astype(y.dtype)
        a_lambdas_flat = A_lambdas.T.ravel().astype(y.dtype)
        b_coeffs_flat = B_coeffs.T.ravel().astype(y.dtype)
        b_lambdas_flat = B_lambdas.T.ravel().astype(y.dtype)
        r_values_flat = R_values.T.ravel().astype(y.dtype)

        # Compute all 6 element values as per SIF structure:
        # A(I,J) = C01 * AC(I,J) * AL(I,J)^(2*ALPHA1-1)
        # B(I,J) = C01 * BC(I,J) * BL(I,J)^(2*ALPHA1-1)
        # C(I,J) = C02 * AC(I,J) * AL(I,J)^(2*ALPHA2-1)
        # D(I,J) = C02 * BC(I,J) * BL(I,J)^(2*ALPHA2-1)
        # E(I,J) = C03 * AC(I,J) * AL(I,J)^(2*ALPHA3-1)
        # F(I,J) = C03 * BC(I,J) * BL(I,J)^(2*ALPHA3-1)
        a_values = c01 * a_coeffs_flat * (a_lambdas_flat**beta1)
        b_values = c01 * b_coeffs_flat * (b_lambdas_flat**beta1)
        c_values = c02 * a_coeffs_flat * (a_lambdas_flat**beta2)
        d_values = c02 * b_coeffs_flat * (b_lambdas_flat**beta2)
        e_values = c03 * a_coeffs_flat * (a_lambdas_flat**beta3)
        f_values = c03 * b_coeffs_flat * (b_lambdas_flat**beta3)

        # Sum all 6 element functions for each (i,j)
        model_values = a_values + b_values + c_values + d_values + e_values + f_values

        # Compute residuals
        residuals = model_values - r_values_flat

        # Return the sum of squared residuals (least squares objective)
        return jnp.sum(residuals**2)

    @property
    def y0(self):
        # Starting point from the SIF file
        return jnp.array([-4.0, -0.1, 4.0, 0.1, -2.0, -0.1])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # The exact solution is not specified in the SIF file
        return None

    @property
    def expected_objective_value(self):
        # The minimum objective value is not precisely specified in the SIF file
        # But the lower bound is given as 0.0
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class SIMPLLPA(AbstractConstrainedMinimisation):
    """A simple linear programming problem in 2 variables.

    Source:
    N. Gould, private communication.

    SIF input: N. Gould, Dec 1989.

    classification: LLR2-AN-2-2
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 2  # 2 variables
    m_eq: int = 0  # no equality constraints
    m_ineq: int = 2  # 2 inequality constraints

    @property
    def y0(self):
        # Both variables start at 0.1
        return jnp.array([0.1, 0.1])

    @property
    def args(self):
        return ()

    def objective(self, y, args):
        """Linear objective function: 2*x1 + x2."""
        x1, x2 = y
        return 2.0 * x1 + x2

    @property
    def bounds(self):
        """Standard LP non-negativity bounds: x >= 0."""
        # Both variables are non-negative
        lower = jnp.zeros(self.n)
        upper = jnp.full(self.n, jnp.inf)
        return lower, upper

    def constraint(self, y):
        """Returns the constraints on the variable y.

        Inequality constraints (g(x) >= 0):
        - x1 + x2 >= 1.0  =>  x1 + x2 - 1.0 >= 0
        - x1 + 2*x2 >= 1.5  =>  x1 + 2*x2 - 1.5 >= 0
        """
        x1, x2 = y

        # No equality constraints
        eq_constraints = None

        # Inequality constraints (must be >= 0)
        g1 = x1 + x2 - 1.0
        g2 = x1 + 2.0 * x2 - 1.5
        ineq_constraints = jnp.array([g1, g2])

        return eq_constraints, ineq_constraints

    @property
    def expected_result(self):
        # The optimal solution is not explicitly given in the SIF file
        return None

    @property
    def expected_objective_value(self):
        # From the SIF file comment
        return jnp.array(1.0)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class POLAK2(AbstractConstrainedMinimisation):
    """POLAK2 problem - A nonlinear minmax problem.

    A nonlinear minmax problem in ten variables.

    Source:
    E. Polak, D.H. Mayne and J.E. Higgins,
    "Superlinearly convergent algorithm for min-max problems"
    JOTA 69, pp. 407-439, 1991.

    SIF input: Ph. Toint, Nov 1993.

    Classification: LOR2-AN-11-2
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self):
        """Number of variables."""
        return 11  # x1, ..., x10, u

    @property
    def m(self):
        """Number of constraints."""
        return 2  # Two inequality constraints

    def objective(self, y, args):
        """Compute the objective (minimize u)."""
        del args
        u = y[10]
        return u

    def constraint(self, y):
        """Compute the constraints."""
        x = y[:10]
        u = y[10]

        # Compute the exponential terms
        def compute_exp_term(x, p):
            a = 1.0e-8 * x[0] ** 2 + (x[1] + p) ** 2
            a = a + x[2] ** 2 + 4.0 * x[3] ** 2
            a = a + x[4] ** 2 + x[5] ** 2 + x[6] ** 2
            a = a + x[7] ** 2 + x[8] ** 2 + x[9] ** 2
            return jnp.exp(a)

        # Constraints:
        # F1: -u + exp(a) with p=2.0 <= 0
        # F2: -u + exp(a) with p=-2.0 <= 0
        e1 = compute_exp_term(x, 2.0)
        e2 = compute_exp_term(x, -2.0)

        c1 = -u + e1
        c2 = -u + e2

        # Both are inequality constraints (<=)
        constraints = jnp.array([c1, c2])

        # Return as inequality constraints
        return None, constraints

    @property
    def y0(self):
        """Initial guess."""
        x0 = jnp.ones(11) * 0.1
        x0 = x0.at[0].set(100.0)  # x1 = 100.0
        return x0

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def bounds(self):
        """Variable bounds (all free)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(54.598146)



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class KOWOSBNE(AbstractNonlinearEquations):
    """A problem arising in the analysis of kinetic data for an enzyme
    reaction, known under the name of Kowalik and Osborne problem
    in 4 variables. This is a nonlinear equation version
    of problem KOWOSB.

    Source:  Problem 15 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    SIF input: Ph. Toint, Dec 1989.
    Modification as a set of nonlinear equations: Nick Gould, Oct 2015.

    classification NOR2-MN-4-11
    """

    n: int = 4
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def num_residuals(self) -> int:
        """Number of residuals."""
        return 11

    def starting_point(self) -> Array:
        """Return the starting point for the problem."""
        return jnp.array([0.25, 0.39, 0.415, 0.39])

    def residual(self, y: Array, args) -> Array:
        """Compute the residual vector.

        Args:
            y: Array of shape (4,) containing [x1, x2, x3, x4]
            args: Additional arguments (unused)

        Returns:
            Array of shape (11,) containing the residuals
        """
        x1, x2, x3, x4 = y[0], y[1], y[2], y[3]

        # U values for each equation
        u = jnp.array(
            [4.0, 2.0, 1.0, 0.5, 0.25, 0.167, 0.125, 0.1, 0.0833, 0.0714, 0.0624]
        )

        # Target values (y-values)
        y_data = jnp.array(
            [
                0.1957,
                0.1947,
                0.1735,
                0.1600,
                0.0844,
                0.0627,
                0.0456,
                0.0342,
                0.0323,
                0.0235,
                0.0246,
            ]
        )

        # Compute the model: x1 * (u^2 + u*x2) / (u^2 + u*x3 + x4)
        u_sq = u * u
        b1 = u_sq + u * x2
        b2 = u_sq + u * x3 + x4
        model = x1 * b1 / b2

        # Residuals
        residuals = model - y_data

        return residuals

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # From the literature, the solution is approximately:
        return jnp.array([0.192807, 0.191282, 0.123057, 0.136062])

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractNonlinearEquations


class HYDCAR6(AbstractNonlinearEquations):
    """The hydrocarbon-6 problem by Fletcher.

    TODO: Human review needed
    Attempts made: Scale factor inversion for NLE
    Suspected issues: Complex scaling interactions, may need element-wise scaling
    Resources needed: Detailed understanding of HYDCAR scaling in pycutest

    Source: Problem 2a in
    J.J. More',"A collection of nonlinear model problems"
    Proceedings of the AMS-SIAM Summer Seminar on the Computational
    Solution of Nonlinear Systems of Equations, Colorado, 1988.
    Argonne National Laboratory MCS-P60-0289, 1989.

    SIF input : N. Gould, Dec 1989

    Classification: NOR2-AN-29-29
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem parameters
    N: int = 6
    M: int = 3
    K: int = 2

    # Constants from SIF file
    A = jnp.array([9.647, 9.953, 9.466])
    B_param = jnp.array([-2998.00, -3448.10, -3347.25])
    C = jnp.array([230.66, 235.88, 215.31])

    AL = jnp.array([0.0, 0.0, 0.0])
    AL_prime = jnp.array([37.6, 48.2, 45.4])
    AL_double_prime = jnp.array([0.0, 0.0, 0.0])

    BE = jnp.array([8425.0, 9395.0, 10466.0])
    BE_prime = jnp.array([24.2, 35.6, 31.9])
    BE_double_prime = jnp.array([0.0, 0.0, 0.0])

    FL = jnp.array([30.0, 30.0, 40.0])
    FV = jnp.array([0.0, 0.0, 0.0])

    TF = 100.0
    B_const = 40.0
    D = 60.0
    Q = 2500000.0

    @property
    def n(self):
        """Number of variables: T(0..5), X(0..5,1..3), V(0..4)"""
        return self.N + self.N * self.M + (self.N - 1)

    def num_residuals(self):
        """Number of residual equations."""
        return 29  # From classification

    def residual(self, y, args):
        """Compute the residuals for the hydrocarbon equations."""
        del args

        # Extract variables - pycutest interleaves T and X by row
        # Pattern: T(0), X(0,1), X(0,2), X(0,3), T(1), X(1,1), ...
        T = jnp.zeros(self.N)
        X = jnp.zeros((self.N, self.M))

        idx = 0
        for i in range(self.N):
            T = T.at[i].set(y[idx])
            idx += 1
            for j in range(self.M):
                X = X.at[i, j].set(y[idx])
                idx += 1

        # V(0..4) come after all T and X
        V = y[idx:]

        # Precompute PI values
        PI = jnp.ones(self.N)
        INVPI = 1.0 / PI

        residuals = []

        # Equations 2.1-(J) for J=1,2,3
        for j in range(self.M):
            eq = self.B_const * X[0, j] - V[0] * X[1, j]
            eq += (
                V[0]
                * X[0, j]
                * INVPI[0]
                * jnp.exp(self.A[j] + self.B_param[j] / (T[0] + self.C[j]))
            )
            residuals.append(eq * 0.01)  # pycutest inverts scale 100

        # Equations 2.2-(I,J) for I=1..4, J=1,2,3
        for i in range(1, self.N - 1):
            for j in range(self.M):
                if i < self.K:
                    p2_21 = self.B_const
                    p2_23 = self.B_const
                elif i == self.K:
                    p2_21 = -self.D
                    p2_23 = self.B_const
                else:
                    p2_21 = -self.D
                    p2_23 = -self.D

                eq = -V[i] * X[i + 1, j] + p2_21 * V[i]
                eq -= (
                    V[i - 1]
                    * X[i - 1, j]
                    * INVPI[i - 1]
                    * jnp.exp(self.A[j] + self.B_param[j] / (T[i - 1] + self.C[j]))
                )
                eq += V[i - 1] * X[i, j] + p2_23 * V[i - 1]
                eq += (
                    V[i]
                    * X[i, j]
                    * INVPI[i]
                    * jnp.exp(self.A[j] + self.B_param[j] / (T[i] + self.C[j]))
                )
                residuals.append(eq * 0.01)  # pycutest inverts scale 100

        # Equations 2.3-(J) for J=1,2,3
        for j in range(self.M):
            eq = -X[self.N - 1, j]
            eq += (
                X[self.N - 2, j]
                * INVPI[self.N - 2]
                * jnp.exp(self.A[j] + self.B_param[j] / (T[self.N - 2] + self.C[j]))
            )
            residuals.append(eq)

        # Equations 2.7-(I) for I=0..5
        for i in range(self.N):
            eq = -1.0
            for j in range(self.M):
                eq += (
                    X[i, j]
                    * INVPI[i]
                    * jnp.exp(self.A[j] + self.B_param[j] / (T[i] + self.C[j]))
                )
            residuals.append(eq)

        # Equation 2.8
        eq = -self.Q
        # Compute SMALLHF and BIGHF
        SMALLHF = 0.0
        BIGHF = 0.0
        for j in range(self.M):
            TEMP1 = self.TF * self.TF * self.AL_double_prime[j]
            TEMP2 = self.TF * self.AL_prime[j]
            TEMP1 = TEMP1 + TEMP2 + self.AL[j]
            TEMP1 = TEMP1 * self.FL[j]
            SMALLHF += TEMP1

            TEMP1 = self.TF * self.TF * self.BE_double_prime[j]
            TEMP2 = self.TF * self.BE_prime[j]
            TEMP1 = TEMP1 + TEMP2 + self.BE[j]
            TEMP1 = TEMP1 * self.FV[j]
            BIGHF += TEMP1

        # E81-(J), E82-(J) and E83-(J) elements
        for j in range(self.M):
            # E81-(J): EXP4PROD
            poly_be = (
                self.BE[j]
                + self.BE_prime[j] * T[0]
                + self.BE_double_prime[j] * T[0] * T[0]
            )
            exprod = INVPI[0] * jnp.exp(
                self.A[j] + self.B_param[j] / (T[0] + self.C[j])
            )
            term = V[0] * X[0, j] * exprod * poly_be
            eq += term

            # E82-(J): POLY1PRD
            poly_al = (
                self.AL[j]
                + self.AL_prime[j] * T[0]
                + self.AL_double_prime[j] * T[0] * T[0]
            )
            term = self.B_const * X[0, j] * poly_al
            eq += term

            # E83-(J): POLY2PRD
            poly_al = (
                self.AL[j]
                + self.AL_prime[j] * T[1]
                + self.AL_double_prime[j] * T[1] * T[1]
            )
            term = -1.0 * X[1, j] * (self.B_const + V[0]) * poly_al
            eq += term

        residuals.append(eq * 1e5)  # Scaled by 1e5

        # Equations 2.9-(I) for I=1..4
        for i in range(1, self.N - 1):
            eq = -SMALLHF if i < self.K else -BIGHF

            for j in range(self.M):
                # E91-(I,J): EXP3PROD
                exprod = INVPI[i - 1] * jnp.exp(
                    self.A[j] + self.B_param[j] / (T[i - 1] + self.C[j])
                )
                term = V[i - 1] * X[i - 1, j] * exprod
                poly = (
                    self.AL[j]
                    + self.AL_prime[j] * T[i - 1]
                    + self.AL_double_prime[j] * T[i - 1] * T[i - 1]
                )
                eq += poly * term

                # E92-(I,J): POLY2PRD
                poly = (
                    self.AL[j]
                    + self.AL_prime[j] * T[i - 1]
                    + self.AL_double_prime[j] * T[i - 1] * T[i - 1]
                )
                p2 = (
                    self.B_const
                    if i < self.K
                    else -self.D
                    if i > self.K
                    else self.B_const
                )
                term = -1.0 * X[i - 1, j] * (p2 + V[i - 1]) * poly
                eq += term

                # E93-(I,J): EXP3PROD
                exprod = INVPI[i] * jnp.exp(
                    self.A[j] + self.B_param[j] / (T[i] + self.C[j])
                )
                term = V[i] * X[i, j] * exprod
                poly = (
                    self.AL[j]
                    + self.AL_prime[j] * T[i]
                    + self.AL_double_prime[j] * T[i] * T[i]
                )
                eq += poly * term

                # E94-(I,J): POLY2PRD
                poly = (
                    self.AL[j]
                    + self.AL_prime[j] * T[i]
                    + self.AL_double_prime[j] * T[i] * T[i]
                )
                p2 = self.B_const if i < self.K else -self.D
                term = -1.0 * X[i, j] * (p2 + V[i]) * poly
                eq += term

            residuals.append(eq * 1e-5)  # pycutest inverts scale 1e5

        return jnp.array(residuals)

    @property
    def y0(self):
        """Initial guess."""
        y0 = jnp.zeros(self.n)

        # X values from SIF file
        X_init = jnp.array(
            [
                [0.0, 0.2, 0.9],
                [0.0, 0.2, 0.8],
                [0.05, 0.3, 0.8],
                [0.1, 0.3, 0.6],
                [0.3, 0.5, 0.3],
                [0.6, 0.6, 0.0],
            ]
        )

        # Interleave T and X values
        idx = 0
        for i in range(self.N):
            y0 = y0.at[idx].set(100.0)  # T(i)
            idx += 1
            for j in range(self.M):
                y0 = y0.at[idx].set(X_init[i, j])  # X(i,j)
                idx += 1

        # V values
        for i in range(self.N - 1):
            y0 = y0.at[idx].set(300.0)
            idx += 1

        return y0

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return jnp.zeros(self.n)  # Placeholder

    @property
    def expected_objective_value(self):
        """Expected optimal objective value (0 for constrained formulation)."""
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class GULFNE(AbstractNonlinearEquations):
    """
    The Gulf RD test problem in 3 variables.
    This function  is a nonlinear least squares with 99 groups.  Each
    group has a nonlinear element of exponential type.

    The number of groups may be varied, but must be larger or equal to 3

    Source: problem 11 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Buckley#27
    SIF input: Ph. Toint, Dec 1989.
    Modification as a set of nonlinear equations: Nick Gould, Oct 2015.

    classification NOR2-MN-3-99
    """

    m: int = 99
    n: int = 3
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def starting_point(self) -> Array:
        return jnp.array([5.0, 2.5, 0.15], dtype=jnp.float64)

    def num_residuals(self) -> int:
        return self.m

    def residual(self, y: Array, args) -> Array:
        """Compute the residuals of the Gulf problem"""
        x1, x2, x3 = y[0], y[1], y[2]

        # Initialize residuals
        residuals = jnp.zeros(self.m, dtype=jnp.float64)

        for i in range(self.m):
            # t = (i+1) * 0.01 for 1-based indexing
            t = (i + 1) * 0.01

            # Gulfian function element calculations
            ymv2 = 25.0 + (-50.0 * jnp.log(t)) ** (2.0 / 3.0) - x2
            a = jnp.abs(ymv2) ** x3 / x1
            expma = jnp.exp(-a)

            # G(i) = E(i) - t
            residuals = residuals.at[i].set(expma - t)

        return residuals

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # Solution is not provided in the SIF file
        return self.starting_point()

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class POWELLSG(AbstractUnconstrainedMinimisation):
    """
    POWELLSG problem.

    The extended Powell singular problem.
    This problem is a sum of n/4 sets of four terms, each of which is
    assigned its own group.

    Source:  Problem 13 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    See also Toint#19, Buckley#34 (p.85)

    SIF input: Ph. Toint, Dec 1989.

    classification OUR2-AN-V-0

    TODO: Human review needed
    Attempts made: Standard interpretation of SCALE factor
    Suspected issues: Possible misinterpretation of SCALE semantics or group definitions
    The objective value is off by a factor of ~4.15
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    n: int = 5000

    def __check_init__(self):
        if self.n % 4 != 0:
            raise ValueError("n must be a multiple of 4 for POWELLSG")

    def objective(self, y, args):
        del args

        # Reshape x into groups of 4
        x_reshaped = y.reshape(-1, 4)
        x1 = x_reshaped[:, 0]
        x2 = x_reshaped[:, 1]
        x3 = x_reshaped[:, 2]
        x4 = x_reshaped[:, 3]

        # G(I): (x[i] + 10*x[i+1])^2
        g1 = (x1 + 10.0 * x2) ** 2

        # G(I+1): 0.2 * (x[i+2] - x[i+3])^2
        g2 = 0.2 * (x3 - x4) ** 2

        # G(I+2): (x[i+1] - 2*x[i+2])^4
        g3 = (x2 - 2.0 * x3) ** 4

        # G(I+3): 0.1 * (x[i] - x[i+3])^4
        g4 = 0.1 * (x1 - x4) ** 4

        return jnp.sum(g1 + g2 + g3 + g4)

    @property
    def y0(self):
        # Starting point: (3, -1, 0, 1) repeated
        n = self.n
        y0 = jnp.zeros(n)
        for i in range(0, n, 4):
            y0 = y0.at[i].set(3.0)
            y0 = y0.at[i + 1].set(-1.0)
            y0 = y0.at[i + 2].set(0.0)
            y0 = y0.at[i + 3].set(1.0)
        return y0

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution not provided in detail in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Solution value is 0.0
        return jnp.array(0.0)

    def num_variables(self):
        return self.n



import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation


class PALMER5A(AbstractBoundedMinimisation):
    """A nonlinear least squares problem with bounds arising from chemical kinetics.

    TODO: Human review needed - Chebyshev polynomial calculation incorrect

    model: H-N=C=Se TZVP + MP2
    fitting Y to A0 T_0 + A2 T_2 + A4 T_4 + A6 T_6 + A8 T_8 +
                A10 T_10 + B / ( C + X**2 ), B, C nonnegative.
    where T_i is the i-th (shifted) Chebyshev polynomial

    Source:
    M. Palmer, Edinburgh, private communication.

    SIF input: Nick Gould, 1992.

    classification SBR2-RN-8-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Data points
    M: int = 23

    # X values (radians) - starting from index 12
    X_data = jnp.array(
        [
            0.000000,
            1.570796,
            1.396263,
            1.308997,
            1.221730,
            1.125835,
            1.047198,
            0.872665,
            0.698132,
            0.523599,
            0.349066,
            0.174533,
        ]
    )

    # Y values (KJmol-1)
    Y_data = jnp.array(
        [
            83.57418,
            81.007654,
            18.983286,
            8.051067,
            2.044762,
            0.000000,
            1.170451,
            10.479881,
            25.785001,
            44.126844,
            62.822177,
            77.719674,
        ]
    )

    @property
    def n(self):
        """Number of variables."""
        return 8  # A0, A2, A4, A6, A8, A10, B, C

    def objective(self, y, args):
        """Compute the objective function (least squares)."""
        del args

        # Extract variables
        A0, A2, A4, A6, A8, A10, B, C = y

        # Constants from SIF
        A = -1.0
        B_const = 1.570796  # X13
        DIFF = 2.0

        # Initialize sum of squares
        obj = 0.0

        for i in range(12):  # Data points from 12 to 23 in SIF
            X = self.X_data[i]
            Y_true = self.Y_data[i]

            # Compute shifted Chebyshev polynomials
            # Y = (2*X - A - B) / DIFF = (2*X - (-1) - 1.570796) / 2
            Y = (2.0 * X - A - B_const) / DIFF

            # T0 = 1.0
            T0 = 1.0
            # T1 = Y
            T1 = Y

            # Recurrence relation: T(j) = 2*Y*T(j-1) - T(j-2)
            # T2 = 2*Y*T1 - T0
            T2 = 2.0 * Y * T1 - T0
            # T3 = 2*Y*T2 - T1
            T3 = 2.0 * Y * T2 - T1
            # T4 = 2*Y*T3 - T2
            T4 = 2.0 * Y * T3 - T2
            # T5 = 2*Y*T4 - T3
            T5 = 2.0 * Y * T4 - T3
            # T6 = 2*Y*T5 - T4
            T6 = 2.0 * Y * T5 - T4
            # T7 = 2*Y*T6 - T5
            T7 = 2.0 * Y * T6 - T5
            # T8 = 2*Y*T7 - T6
            T8 = 2.0 * Y * T7 - T6
            # T9 = 2*Y*T8 - T7
            T9 = 2.0 * Y * T8 - T7
            # T10 = 2*Y*T9 - T8
            T10 = 2.0 * Y * T9 - T8

            # Model prediction
            X_sqr = X * X
            prediction = (
                A0 * T0
                + A2 * T2
                + A4 * T4
                + A6 * T6
                + A8 * T8
                + A10 * T10
                + B / (C + X_sqr)
            )

            # Residual
            residual = prediction - Y_true

            # Add to sum of squares
            obj += residual * residual

        return jnp.array(obj)

    @property
    def y0(self):
        """Initial guess."""
        return jnp.ones(8)

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return jnp.zeros(8)  # Placeholder

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        return jnp.array(4.0606141e-02)

    @property
    def bounds(self):
        """Returns the bounds on the variable y."""
        # A0-A10 are free, B and C have lower bounds (B=0, C=0.00001 from pycutest)
        lower = jnp.array(
            [-jnp.inf, -jnp.inf, -jnp.inf, -jnp.inf, -jnp.inf, -jnp.inf, 0.0, 0.00001]
        )
        upper = jnp.full(8, jnp.inf)
        return lower, upper



import json
import os

import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


# Load data at module level to avoid tracer leaks
def _load_haifal_data():
    """Load and preprocess HAIFAL data."""
    data_path = os.path.join(os.path.dirname(__file__), "data", "haifal_data.json")

    with open(data_path) as f:
        data = json.load(f)

    # Load the constraint mapping
    mapping_path = os.path.join(
        os.path.dirname(__file__), "data", "haifal_constraint_mapping.json"
    )

    with open(mapping_path) as f:
        mapping_data = json.load(f)

    # Convert to JAX arrays
    element_x_indices = jnp.array(data["element_x_indices"], dtype=jnp.int32)
    element_y_indices = jnp.array(data["element_y_indices"], dtype=jnp.int32)
    constraint_id_order = (
        jnp.array(mapping_data["constraint_id_order"], dtype=jnp.int32) - 1
    )  # Convert to 0-based

    # Build vectorized sparse representation
    # Flatten all element indices and coefficients
    element_indices_flat = []
    coefficients_flat = []
    constraint_indices_flat = []

    constraint_element_lists = data["constraint_element_lists"]
    constraint_coeff_lists = data["constraint_coeff_lists"]

    for constraint_idx, (elements, coeffs) in enumerate(
        zip(constraint_element_lists, constraint_coeff_lists)
    ):
        if elements:  # Only process non-empty constraints
            constraint_id = constraint_id_order[constraint_idx]
            for elem_idx, coeff in zip(elements, coeffs):
                element_indices_flat.append(elem_idx - 1)  # Convert to 0-based
                coefficients_flat.append(coeff)
                constraint_indices_flat.append(constraint_id)

    # Convert to JAX arrays
    element_indices_flat = jnp.array(element_indices_flat, dtype=jnp.int32)
    coefficients_flat = jnp.array(coefficients_flat)
    constraint_indices_flat = jnp.array(constraint_indices_flat, dtype=jnp.int32)

    return (
        element_x_indices,
        element_y_indices,
        element_indices_flat,
        coefficients_flat,
        constraint_indices_flat,
    )


# Load data once at module level
_HAIFAL_DATA = _load_haifal_data()


class HAIFAL(AbstractConstrainedMinimisation):
    """Truss Topology Design problem HAIFAL (171-8940).

    A truss topology design optimization problem with 343 variables and 8940
    constraints. This is a quadratic minimization problem with inequality
    constraints derived from structural engineering applications.

    Variables:
    - z: objective variable to be minimized
    - x(1), ..., x(342): design variables

    Objective: minimize z

    Constraints: 8940 inequality constraints of the form C(i) - z - x(333) ≤ 0,
    where each C(i) contains quadratic terms in the design variables.

    Source: M. Tsibulevsky, Optimization Laboratory,
    Faculty of Industrial Engineering, Technion,
    Haifa, 32000, Israel.

    SIF input: Conn, Gould and Toint, May, 1992

    Classification: LQR2-AN-343-8940
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        # Minimize z (the first variable)
        return y[0]

    @property
    def y0(self):
        # Starting point: all variables initialized to 0 (matching pycutest)
        return jnp.zeros(343)

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution not provided in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Lower bound is 0.0 according to SIF file
        return None

    @property
    def bounds(self):
        # All variables are real with no explicit bounds
        return None

    def constraint(self, y):
        z = y[0]
        x = y[1:343]  # x[0] corresponds to X(1) in SIF, etc.
        x333 = x[332]  # X(333) - appears in all constraints

        # Use preloaded vectorized data
        (
            element_x_indices,
            element_y_indices,
            element_indices_flat,
            coefficients_flat,
            constraint_indices_flat,
        ) = _HAIFAL_DATA

        # Vectorized element computation: E(i) = 0.5 * X(idx1) * X(idx2)
        # All 74,423 elements computed at once
        element_values = 0.5 * x[element_x_indices] * x[element_y_indices]

        # Initialize all constraints with base formula: -z - x333
        constraints = jnp.full(8958, -z - x333)

        # Fully vectorized constraint computation using scatter-add
        # Select relevant element values and multiply by coefficients
        relevant_elements = element_values[element_indices_flat]
        weighted_contributions = coefficients_flat * relevant_elements

        # Use scatter-add to accumulate contributions to each constraint
        constraints = constraints.at[constraint_indices_flat].add(
            weighted_contributions
        )

        return None, constraints



"""Price problem 3."""

from __future__ import annotations

from typing import Any

import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class PRICE3(AbstractUnconstrainedMinimisation):
    """SCIPY global optimization benchmark example Price03.

    Fit: y = (10(x_1^2-x_2), sqrt(6)(6.4(x_2-0.5)^2 -x_1)) + e

    Source: Problem from the SCIPY benchmark set
    https://github.com/scipy/scipy/tree/master/benchmarks/
    benchmarks/go_benchmark_functions

    SIF input: Nick Gould, Jan 2020

    classification SUR2-MN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y: Any, args: Any) -> Any:
        """Compute the objective function."""
        x1, x2 = y

        # Element functions
        # E1: SQR type with x1: x1^2
        e1 = x1 * x1

        # E2: SSQR type with x2: (x2 - 0.5)^2
        e2 = (x2 - 0.5) ** 2

        # Group functions with L2 type (squares the result)
        # F1: (e1 - x2)^2 / 0.01 = (x1^2 - x2)^2 / 0.01 (SCALE means division)
        f1 = (e1 - x2) ** 2 / 0.01

        # F2: (6.4 * e2 - x1 - 0.6)^2
        # Testing against pycutest shows the constant should be subtracted, not added
        # This gives the correct objective value
        f2 = (6.4 * e2 - x1 - 0.6) ** 2

        return f1 + f2

    @property
    def y0(self):
        return jnp.array([1.0, 5.0])

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # The optimal solution should have objective value 0
        # This occurs when both terms are zero:
        # x1^2 = x2 and 6.4*(x2 - 0.5)^2 = x1 - 0.6
        return None

    @property
    def expected_objective_value(self):
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._problem import AbstractBoundedMinimisation


# TODO(claude): move to folder with bounded minimisation problems. Create new subclass
# AbstractBoundedQuadraticProblem in sif2jax/_problem.py. Then test, commit and push.
class QUDLIN(AbstractBoundedMinimisation):
    """A simple quadratic programming problem.

    The objective consists of linear terms and product terms x_i * x_{i+1}.

    SIF input: unknown.
    minor correction by Ph. Shott, Jan 1995.

    classification QBR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    n: int = 5000  # Number of variables
    m: int = 2500  # Number of product terms

    @property
    def y0(self):
        """Initial guess - zeros."""
        return jnp.zeros(self.n)

    @property
    def args(self):
        return None

    def objective(self, y, args):
        """Quadratic objective function.

        Linear terms: -10 * i * x_i for i=1 to n
        Quadratic terms: x_i * x_{i+1} for i=1 to m
        """
        del args

        # Linear terms - the SIF file has RM C RI -10.0 which means C = RI * (-10.0)
        i_vals = jnp.arange(1, self.n + 1)
        i_vals = jnp.asarray(i_vals, dtype=y.dtype)
        coeffs = -10.0 * i_vals
        linear_term = jnp.dot(coeffs, y)

        # Quadratic terms: sum of x_i * x_{i+1} for i=1 to m
        # From the SIF file: ELEMENT uses 2PR type which computes X*Y
        quad_term = 0.0
        if self.m > 0 and self.n > 1:
            # Ensure we don't exceed array bounds
            max_idx = min(self.m, self.n - 1)
            quad_term = jnp.sum(y[:max_idx] * y[1 : max_idx + 1])

        return linear_term + quad_term

    @property
    def bounds(self):
        """Variable bounds: 0 <= x_i <= 10."""
        lower = jnp.zeros(self.n)
        upper = jnp.full(self.n, 10.0)
        return lower, upper

    @property
    def expected_result(self):
        """Expected result not provided in SIF file."""
        return None

    @property
    def expected_objective_value(self):
        """Expected objective value is 0.0 (from SIF file)."""
        return jnp.array(0.0)



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class ARGTRIG(AbstractNonlinearEquations):
    """
    Variable dimension trigonometric problem
    This problem is a sum of n least-squares groups, each of
    which has n+1 nonlinear elements.  Its Hessian matrix is dense.

    Source:  Problem 26 in
    J.J. More', B.S. Garbow and K.E. Hillstrom,
    "Testing Unconstrained Optimization Software",
    ACM Transactions on Mathematical Software, vol. 7(1), pp. 17-41, 1981.

    SIF input: Ph. Toint, Dec 1989.

    classification NOR2-AN-V-V
    """

    n: int = 200
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def starting_point(self) -> Array:
        return jnp.full(self.n, 1.0 / self.n, dtype=jnp.float64)

    def num_residuals(self) -> int:
        return self.n

    def residual(self, y: Array, args) -> Array:
        """Compute the residuals of the trigonometric problem"""
        n = self.n

        # Compute residuals
        residuals = []
        for i in range(n):
            # For each group i, compute:
            # REALI * (cos(y[i]) + sin(y[i])) + sum(cos(y[j]) for j in range(n))
            # minus (n + i + 1)
            sum_cos = jnp.sum(jnp.cos(y))
            res_i = (
                float(i + 1) * (jnp.cos(y[i]) + jnp.sin(y[i]))
                + sum_cos
                - float(n + i + 1)
            )
            residuals.append(res_i)

        return jnp.array(residuals)

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> None:
        """Expected result of the optimization problem."""
        # The SIF file mentions solution value 0.0, but not the exact solution vector
        return None

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """No bounds for this problem."""
        return None



from jax import numpy as jnp
from jaxtyping import Array, Float

from ..._problem import AbstractNonlinearEquations


class ARTIF(AbstractNonlinearEquations):
    """An artificial nonlinear system.

    Source:
    K.M. Irani, M.P. Kamat, C.J. Ribbens, H.F.Walker and L.T. Watson,
    "Experiments with conjugate gradient algorithms for homotopy curve
     tracking" ,
    SIAM Journal on Optimization, May 1991, pp. 222-251, 1991.

    SIF input: Ph. Toint, May 1990.

    classification NOR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})
    n: int = 5002  # Total number of variables including fixed boundary values
    n_eq: int = 5000  # Number of equations

    def residual(self, y, args) -> Float[Array, "5000"]:
        """Residual function for the nonlinear equations."""
        # y already contains all variables including fixed boundary values
        x = y

        # Vectorized computation
        # Create index arrays
        i = jnp.arange(1, self.n_eq + 1)

        # Linear part: -0.05 * (X(i-1) + X(i) + X(i+1))
        linear_part = -0.05 * (x[i - 1] + x[i] + x[i + 1])

        # Nonlinear part: arctan(sin(fact * X(i)))
        # fact = i % 100
        fact = (i % 100).astype(x.dtype)
        nonlinear_part = jnp.arctan(jnp.sin(fact * x[i]))

        return linear_part + nonlinear_part

    @property
    def y0(self) -> Float[Array, "5002"]:
        """Initial guess for the optimization problem."""
        # All variables start at 1.0, including fixed boundary values
        # Note: pycutest keeps boundary values at 1.0 in initial guess
        # even though they are fixed to 0.0 in bounds
        return jnp.ones(self.n)

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> None:
        """Expected result of the optimization problem."""
        # The SIF file doesn't provide a solution
        return None

    @property
    def expected_objective_value(self) -> Float[Array, ""]:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """Bounds for the variables."""
        # X(0) and X(N+1) are fixed at 0
        lower = jnp.full(self.n, -jnp.inf)
        upper = jnp.full(self.n, jnp.inf)

        # Fix boundary values
        lower = lower.at[0].set(0.0)
        upper = upper.at[0].set(0.0)
        lower = lower.at[-1].set(0.0)
        upper = upper.at[-1].set(0.0)

        return lower, upper



r"""
The "ten bar truss" structural optimization problem, version P2.

The problem is to minimize the cross section areas of the bars
in the structure

    /|
    /|>o------o------o
    /|  \    /|\    /|
         \  / | \  / |
          \/  |  \/  |
          /\  |  /\  |
         /  \ | /  \ |
    /|  /    \|/    \|
    /|>o------o------o
    /|

submitted to vertical forces of equal magnitude (P0) applied at
the two free lower nodes, subject to limits of nodal displacements.

Source:
K. Svanberg,
private communication,  August 1990.
See also
K. Svanberg,
"On local and global minima in structural optimization",
in "New directions in optimum structural design" (Atrek, Ragsdell
and Zienkiwewicz, eds.), Wiley, 1984.

SIF input: Ph. Toint, August 1990.

classification LOR2-MY-18-8
"""

import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class TENBARS2(AbstractConstrainedMinimisation):
    @property
    def name(self) -> str:
        return "TENBARS2"

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 18  # 8 nodal displacements (U1-U8) + 10 bar cross sections (X1-X10)
    n_equality_constraints: int = 8  # 8 equilibrium conditions
    n_inequality_constraints: int = 0  # No inequality constraints

    @property
    def y0(self):
        # All variables start at 0.0 according to SIF file
        return jnp.zeros(18)

    @property
    def args(self):
        return None

    def objective(self, y, args):
        del args
        # Extract cross section areas (last 10 variables)
        x = y[8:]  # X1 to X10

        # Constants
        c0 = 2.53106
        sq2 = jnp.sqrt(2.0)
        c0sq2 = c0 * sq2

        # Objective function is sum of cross section areas times their factors
        obj = c0 * (x[0] + x[3] + x[4] + x[5] + x[8] + x[9])  # X1, X4, X5, X6, X9, X10
        obj += c0sq2 * (x[1] + x[2] + x[6] + x[7])  # X2, X3, X7, X8

        return jnp.array(obj)

    def constraint(self, y):
        # Extract nodal displacements and cross sections
        u = y[:8]  # U1 to U8
        x = y[8:]  # X1 to X10

        # Constants
        sq8 = jnp.sqrt(8.0)
        inv_sq8 = 1.0 / sq8
        minus_inv_sq8 = -inv_sq8
        p0 = 589.884  # -P0 from SIF file (C4 and C8 RHS values)

        # Element computations
        # EA: X1 * U1
        ea = x[0] * u[0]

        # EB: X3 * (U1 + U2)
        eb = x[2] * (u[0] + u[1])

        # EC: X6 * (U1 - U5)
        ec = x[5] * (u[0] - u[4])

        # ED: X7 * (U1 + U8 - U2 - U7)
        ed = x[6] * (u[0] + u[7] - u[1] - u[6])

        # EE: X5 * (U2 - U4)
        ee = x[4] * (u[1] - u[3])

        # EF: X2 * (U3 - U4) - corrected based on TENBARS1 analysis
        # The SIF file may specify (U3 + U4), but pycutest uses (U3 - U4)
        ef = x[1] * (u[2] - u[3])

        # EG: X4 * U3
        eg = x[3] * u[2]

        # EH: X8 * (U3 + U4 - U5 - U6)
        eh = x[7] * (u[2] + u[3] - u[4] - u[5])

        # EI: X9 * (U3 - U7)
        ei = x[8] * (u[2] - u[6])

        # EJ: X10 * (U6 - U8)
        ej = x[9] * (u[5] - u[7])

        # Equality constraints (C1 to C8)
        c1 = ea + ec + inv_sq8 * eb + inv_sq8 * ed
        c2 = inv_sq8 * eb + ee + minus_inv_sq8 * ed
        # EF term: X2*(U3-U4) added to C3, subtracted from C4
        c3 = inv_sq8 * eh + ei + eg + inv_sq8 * ef
        c4 = inv_sq8 * eh - ee + p0 - inv_sq8 * ef
        c5 = minus_inv_sq8 * eh - ec
        c6 = minus_inv_sq8 * eh + ej
        c7 = minus_inv_sq8 * ed - ei
        c8 = inv_sq8 * ed - ej + p0

        equality_constraints = jnp.array([c1, c2, c3, c4, c5, c6, c7, c8])

        # No inequality constraints for TENBARS2
        inequality_constraints = None

        return equality_constraints, inequality_constraints

    @property
    def bounds(self):
        # Lower bounds on U (even indices 2,4,6,8 have -50.8 lower bound)
        # and on cross sections (0.645)
        lower = jnp.concatenate(
            [
                jnp.array(
                    [
                        -jnp.inf,  # U1
                        -50.8,  # U2
                        -jnp.inf,  # U3
                        -50.8,  # U4
                        -jnp.inf,  # U5
                        -50.8,  # U6
                        -jnp.inf,  # U7
                        -50.8,  # U8
                    ]
                ),
                jnp.full(10, 0.645),  # Lower bound 0.645 on cross sections
            ]
        )
        upper = jnp.full(18, jnp.inf)
        return lower, upper

    @property
    def expected_result(self):
        # The optimal solution is not explicitly given in the SIF file
        return None

    @property
    def expected_objective_value(self):
        # According to the SIF file comments, two solution values are given:
        # SOLTN-A: 2302.55
        # SOLTN-B: 2277.9458
        # We'll use the B solution as the expected value
        return jnp.array(2277.9458)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class DTOC6(AbstractConstrainedMinimisation):
    """A discrete time optimal control (DTOC) problem with nonlinear objective.

    The system has N time periods, 1 control variable and 1 state variable.

    The problem is convex.

    Sources: problem 6 in
    T.F. Coleman and A. Liao,
    "An Efficient Trust Region Method for Unconstrained Discrete-Time Optimal
    Control Problems",
    Tech. Report, ctc93tr144,  Advanced Computing Research Institute,
    Cornell University, 1992.

    D.M. Murray and S.J. Yakowitz,
    "The application of optimal control methodology to nonlinear programming
    problems",
    Mathematical Programming 21, pp. 331-347, 1981.

    SIF input: Ph. Toint, August 1993

    classification OOR2-AN-V-V
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem parameters - using the default values from SIF
    n_periods: int = 5001  # N
    n_controls: int = 1  # NX=1
    n_states: int = 1  # NY=1

    # Total number of variables and constraints
    n: int = (n_periods - 1) * n_controls + n_periods * n_states  # 10001
    m: int = n_periods - 1  # 5000

    @property
    def y0(self):
        # Start from zeros except for Y(1) = 0.0 (which is already 0)
        return jnp.zeros(self.n)

    @property
    def args(self):
        # No special args needed for this problem
        return None

    def objective(self, y, args):
        # Extract control and state variables
        # Note: We don't enforce fixed variables here as pycutest evaluates
        # with the values given in the input vector
        # X(t) for t=1..N-1
        # Y(t) for t=1..N
        x_vars = y[: self.n_periods - 1]
        y_vars = y[self.n_periods - 1 :]

        # Objective function: sum_{t=1}^{N-1} 2*(Y(t)^2 + exp(X(t))*Y(t) + X(t)^2)
        # Note: sum is over t=1..N-1, so we use Y(1)..Y(N-1)
        y_obj = y_vars[:-1]  # Y(t) for t=1..N-1

        # Compute exp(X(t))
        exp_x = jnp.exp(x_vars)

        # The objective groups OY(t) and OX(t) both use L2 group type with scale 2.0
        # L2 group type squares the GVAR
        # OY(t): 2.0 * (Y(t) + exp(X(t)))^2
        # OX(t): 2.0 * X(t)^2
        # Total objective: sum of OY(t) + OX(t)

        # OY(t) group: GVAR = Y(t), nonlinear element E(T) = exp(X(t))
        # L2 group: (GVAR + E(T))^2 * scale
        # The SIF file says scale = 2.0 but pycutest interprets this as dividing by 2
        oy_terms = 0.5 * (y_obj + exp_x) ** 2

        # OX(t) group: GVAR = X(t), no nonlinear element
        # L2 group: GVAR^2 * scale
        # The SIF file says scale = 2.0 but pycutest interprets this as dividing by 2
        ox_terms = 0.5 * x_vars**2

        obj_sum = jnp.sum(oy_terms + ox_terms)

        return obj_sum

    def constraint(self, y):
        # Extract control and state variables
        # Note: We don't enforce fixed variables here as pycutest evaluates
        # with the values given in the input vector
        x_vars = y[: self.n_periods - 1]
        y_vars = y[self.n_periods - 1 :]

        # Transition constraints with exponential term
        # TT(t): -Y(t+1) + Y(t) + exp(X(t)) = 0
        # for t=1..N-1

        y_current = y_vars[:-1]  # Y(t) for t=1..N-1
        y_next = y_vars[1:]  # Y(t+1) for t=1..N-1

        # Compute exp(X(t))
        exp_x = jnp.exp(x_vars)

        # Compute constraints
        constraints = -y_next + y_current + exp_x

        return constraints, None

    @property
    def bounds(self):
        # Y(1)=0.0 is fixed, others are free
        n_x = self.n_periods - 1

        lbs = jnp.full(self.n, -jnp.inf)
        ubs = jnp.full(self.n, jnp.inf)

        # Y(1) at index n_x is fixed to 0.0
        lbs = lbs.at[n_x].set(0.0)
        ubs = ubs.at[n_x].set(0.0)

        return lbs, ubs

    @property
    def expected_result(self) -> None:
        """Solution not provided in SIF file"""
        return None

    @property
    def expected_objective_value(self) -> None:
        """Optimal value for N=5001 not provided in SIF file"""
        return None



import jax.numpy as jnp
from jax import Array

from ..._problem import AbstractNonlinearEquations


class DENSCHNCNE(AbstractNonlinearEquations):
    """
    Source: an example problem (p. 98) in
    J.E. Dennis and R.B. Schnabel,
    "Numerical Methods for Unconstrained Optimization and Nonlinear
    Equations",
    Prentice-Hall, Englewood Cliffs, 1983.

    SIF input: Ph. Toint, Nov 1990.
    Nonlinear-equations version of DENSCHNC.SIF, Nick Gould, Jan 2020.

    classification NOR2-AN-2-2
    """

    n: int = 2
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def starting_point(self) -> Array:
        return jnp.array([2.0, 3.0], dtype=jnp.float64)

    def num_residuals(self) -> int:
        return 2

    def residual(self, y: Array, args) -> Array:
        """Compute the residuals of the Dennis-Schnabel problem C"""
        x1, x2 = y[0], y[1]

        # Group A: x1^2 + x2^2 - 2
        res_a = x1**2 + x2**2 - 2.0

        # Group B: exp(x1 - 1) + x2^3 - 2
        res_b = jnp.exp(x1 - 1.0) + x2**3 - 2.0

        return jnp.array([res_a, res_b])

    @property
    def y0(self) -> Array:
        """Initial guess for the optimization problem."""
        return self.starting_point()

    @property
    def args(self):
        """Additional arguments for the residual function."""
        return None

    @property
    def expected_result(self) -> Array:
        """Expected result of the optimization problem."""
        # Solution is not provided in the SIF file
        return self.starting_point()

    @property
    def expected_objective_value(self) -> Array:
        """Expected value of the objective at the solution."""
        # For nonlinear equations with pycutest formulation, this is always zero
        return jnp.array(0.0)

    def constraint(self, y):
        """Returns the residuals as equality constraints."""
        return self.residual(y, self.args), None

    @property
    def bounds(self) -> tuple[Array, Array] | None:
        """No bounds for this problem."""
        return None



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class HAIFAS(AbstractConstrainedMinimisation):
    """Truss Topology Design problem HAIFAS (t6-9).

    A truss topology design optimization problem with 13 variables and 9 constraints.
    This is a quadratic minimization problem with inequality constraints derived
    from structural engineering applications.

    Variables:
    - z: objective variable to be minimized
    - x(1), ..., x(12): design variables

    Objective: minimize z

    Constraints: 9 inequality constraints of the form G(i) - z - x(10) ≤ 0,
    where each G(i) contains quadratic terms in the design variables.

    Source: M. Tsibulevsky, Optimization Laboratory,
    Faculty of Industrial Engineering, Technion,
    Haifa, 32000, Israel.

    SIF input: Conn, Gould and Toint, May, 1992
    minor correction by Ph. Shott, Jan 1995.

    Classification: LQR2-AN-13-9
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        # Minimize z (the first variable)
        return y[0]

    @property
    def y0(self):
        # Starting point: all variables initialized to 0
        return jnp.zeros(13)

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution not provided in SIF file
        return None

    @property
    def expected_objective_value(self):
        # Lower bound is 0.0 according to SIF file
        return None

    @property
    def bounds(self):
        # All variables are real with no explicit bounds
        return None

    def constraint(self, y):
        z = y[0]
        x = y[1:13]  # x[0] corresponds to X(1) in SIF, etc.

        # Extract individual variables (0-indexed)
        x4, x5, x6 = x[3], x[4], x[5]  # X(4), X(5), X(6)
        x10, x11, x12 = x[9], x[10], x[11]  # X(10), X(11), X(12)

        # Constraint functions G(i) based on the SIF analysis
        g1 = 5.0 * x4**2

        g2 = 3.2 * x5**2 + 3.2 * x5 * x11 + 0.8 * x11**2

        g3 = 20.0 * x10**2 - 40.0 * x10 * x11 + 20.0 * x11**2

        g4 = 3.2 * x4**2 - 3.2 * x4 * x10 + 0.8 * x10**2

        g5 = 5.0 * x5**2

        g6 = 3.2 * x6**2 + 3.2 * x6 * x12 + 0.8 * x12**2

        g7 = 20.0 * x11**2 - 40.0 * x11 * x12 + 20.0 * x12**2

        g8 = 3.2 * x5**2 - 3.2 * x5 * x11 + 0.8 * x11**2

        g9 = 5.0 * x6**2

        # Each constraint: G(i) - z - x(10) ≤ 0
        # Rearranged as: G(i) - z - x(10)
        inequality_constraints = jnp.array(
            [
                g1 - z - x10,
                g2 - z - x10,
                g3 - z - x10,
                g4 - z - x10,
                g5 - z - x10,
                g6 - z - x10,
                g7 - z - x10,
                g8 - z - x10,
                g9 - z - x10,
            ]
        )

        return None, inequality_constraints



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


class FCCU(AbstractConstrainedMinimisation):
    """FCCU - Fluid catalytic cracker data reconciliation problem.

    A data reconciliation problem for a fluid catalytic cracker (FCCU) that
    finds computed flows that best match measured flows while satisfying
    mass balance constraints.

    Variables: 19 flow rates through the cracker system
    Objective: minimize ∑ᵢ Wᵢ(Cᵢ - Mᵢ)² (weighted least squares)
    Constraints: 8 linear mass balance equations at nodes F1-F8

    Where:
    - Cᵢ = computed flow (variable)
    - Mᵢ = measured flow (constant)
    - Wᵢ = weight for flow i

    Flow variables:
    Feed, Effluent, MF_ohd, HCN, LCO, HCO, MF_btms, Decant, Dec_recy,
    Off_gas, DC4_feed, DC3_feed, DC4_btms, Lean_oil, Propane, Butane,
    C8spl_fd, LCN, MCN

    Source: W. J. Korchinski, Profimatics, Inc, Spring 1993.

    Classification: SLR2-MN-19-8
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def y0(self):
        # Starting point from SIF: all variables = 1.0
        return jnp.array([1.0] * 19)

    @property
    def args(self):
        return None

    def objective(self, y, args):
        del args

        # Extract flow variables
        (
            Feed,
            Effluent,
            MF_ohd,
            HCN,
            LCO,
            HCO,
            MF_btms,
            Decant,
            Dec_recy,
            Off_gas,
            DC4_feed,
            DC3_feed,
            DC4_btms,
            Lean_oil,
            Propane,
            Butane,
            C8spl_fd,
            LCN,
            MCN,
        ) = y

        # Weights for objective function (using exact fractions for precision)
        weights = jnp.array(
            [
                0.2,
                1.0,
                1.0,
                1.0 / 3.0,  # More precise than 0.33333333
                1.0 / 3.0,
                1.0 / 3.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0 / 3.0,
                1.0 / 3.0,
                1.0,
                1.0 / 3.0,
                1.0 / 3.0,
            ]
        )

        # Measured flows (constants)
        measured = jnp.array(
            [
                31.0,
                36.0,
                20.0,
                3.0,
                5.0,
                3.5,
                4.2,
                0.9,
                3.9,
                2.2,
                22.8,
                6.8,
                19.0,
                8.5,
                2.2,
                2.5,
                10.8,
                6.5,
                6.5,
            ]
        )

        # TODO: Human review needed for SIF interpretation
        # Current formulation: W_i * (y_i - M_i)² + W_i * M_i²
        # Attempts made:
        # 1. Basic W_i * (y_i - M_i)² -> off by factor 2.63
        # 2. Various scale factors and constant adjustments
        # 3. Analysis of SIF SQUARE group and CONSTANTS sections
        # Expected pycutest value: 7362.59, current gives: 5891.37
        # Suspected issues: SIF group scaling interpretation, pycutest differences
        # Resources needed: SIF specification docs, data reconciliation examples

        differences = y - measured
        weighted_squares = jnp.sum(weights * differences**2)
        constant_term = jnp.sum(weights * measured**2)
        return weighted_squares + constant_term

    def constraint(self, y):
        # Extract flow variables
        (
            Feed,
            Effluent,
            MF_ohd,
            HCN,
            LCO,
            HCO,
            MF_btms,
            Decant,
            Dec_recy,
            Off_gas,
            DC4_feed,
            DC3_feed,
            DC4_btms,
            Lean_oil,
            Propane,
            Butane,
            C8spl_fd,
            LCN,
            MCN,
        ) = y

        # Linear mass balance constraints (equality)
        # F1: Feed + Dec_recy - Effluent = 0
        f1 = Feed + Dec_recy - Effluent

        # F2: Effluent - MF_ohd - HCN - LCO - HCO - MF_btms = 0
        f2 = Effluent - MF_ohd - HCN - LCO - HCO - MF_btms

        # F3: MF_btms - Decant - Dec_recy = 0
        f3 = MF_btms - Decant - Dec_recy

        # F4: MF_ohd + Lean_oil - Off_gas - DC4_feed = 0
        f4 = MF_ohd + Lean_oil - Off_gas - DC4_feed

        # F5: DC4_feed - DC3_feed - DC4_btms = 0
        f5 = DC4_feed - DC3_feed - DC4_btms

        # F6: DC4_btms - Lean_oil - C8spl_fd = 0
        f6 = DC4_btms - Lean_oil - C8spl_fd

        # F7: DC3_feed - Propane - Butane = 0
        f7 = DC3_feed - Propane - Butane

        # F8: C8spl_fd - LCN - MCN = 0
        f8 = C8spl_fd - LCN - MCN

        eq_constraints = jnp.array([f1, f2, f3, f4, f5, f6, f7, f8])

        # No inequality constraints
        ineq_constraints = None

        return eq_constraints, ineq_constraints

    @property
    def bounds(self):
        # Flow rates should be non-negative (19 variables)
        lower = jnp.zeros(19)
        upper = jnp.full(19, jnp.inf)
        return (lower, upper)

    @property
    def expected_result(self):
        # Optimal solution from SIF file
        return jnp.array(
            [
                3.11639e1,
                3.53528e1,
                1.94669e1,
                2.94255e0,
                4.94255e0,
                3.44255e0,
                4.55828e0,
                3.69371e-1,
                4.18891e0,
                2.56075e0,
                2.41207e1,
                5.15601e0,
                1.89647e1,
                7.21458e0,
                2.42801e0,
                2.72801e0,
                1.17501e1,
                5.87506e0,
                5.87506e0,
            ]
        )

    @property
    def expected_objective_value(self):
        # Optimal objective value from SIF file
        return jnp.array(1.11491e1)



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class SPARSINE(AbstractUnconstrainedMinimisation):
    """A sparse problem involving sine functions.

    This problem has a sparse structure where each objective group includes
    sine elements from specific positions determined by modular arithmetic.

    Source: Nick Gould, November 1995

    Classification: OUR2-AN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 5000  # Number of variables

    def __init__(self, n: int = 5000):
        self.n = n

    def objective(self, y, args):
        del args
        n = self.n

        # Compute sine of all variables
        sine_values = jnp.sin(y)

        # Vectorized computation: for each group OBJ(i),
        # compute the sum of sine elements
        # Group i includes sine elements from positions:
        # i, (2i-1) mod n + 1, (3i-1) mod n + 1, (5i-1) mod n + 1,
        # (7i-1) mod n + 1, (11i-1) mod n + 1

        # Create array of group indices (1 to n) using same dtype as y
        i_values = jnp.arange(1, n + 1, dtype=y.dtype)

        # Compute all position indices in vectorized form (convert to 0-based)
        # SIF formula: (k*i-1) mod n + 1, convert to 0-based by subtracting 1
        # Result: ((k*i-1) mod n + 1) - 1 = (k*i-1) mod n
        pos_i = (i_values - 1).astype(jnp.int32)  # i (0-based)
        pos_2i = ((2 * i_values - 1) % n).astype(jnp.int32)  # (2i-1) mod n
        pos_3i = ((3 * i_values - 1) % n).astype(jnp.int32)  # (3i-1) mod n
        pos_5i = ((5 * i_values - 1) % n).astype(jnp.int32)  # (5i-1) mod n
        pos_7i = ((7 * i_values - 1) % n).astype(jnp.int32)  # (7i-1) mod n
        pos_11i = ((11 * i_values - 1) % n).astype(jnp.int32)  # (11i-1) mod n

        # Gather sine values at all positions
        sine_i = sine_values[pos_i]
        sine_2i = sine_values[pos_2i]
        sine_3i = sine_values[pos_3i]
        sine_5i = sine_values[pos_5i]
        sine_7i = sine_values[pos_7i]
        sine_11i = sine_values[pos_11i]

        # Sum of sine values for each group (alpha values)
        alpha_values = sine_i + sine_2i + sine_3i + sine_5i + sine_7i + sine_11i

        # Group contributions: 0.5 * P * ALPHA^2 where P = i
        group_contributions = 0.5 * i_values * alpha_values**2

        return jnp.sum(group_contributions)

    @property
    def bounds(self):
        # All variables are free (unbounded)
        lower = jnp.full(self.n, -jnp.inf)
        upper = jnp.full(self.n, jnp.inf)
        return lower, upper

    @property
    def y0(self):
        # Starting point: all variables = 0.5
        return jnp.full(self.n, 0.5)

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        # Solution is likely all zeros (minimum of sine functions)
        return jnp.zeros(self.n)

    @property
    def expected_objective_value(self):
        return jnp.array(0.0)



import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractConstrainedQuadraticProblem


class EIGENA2(AbstractConstrainedQuadraticProblem):
    """Solving symmetric eigenvalue problems as constrained quadratic optimization.

    The problem is, given a symmetric matrix A, to find an orthogonal
    matrix Q and diagonal matrix D such that A Q(T) = Q(T) D.

    Example A: a diagonal matrix with eigenvalues 1, ..., N.

    Source: An idea by Nick Gould

    Constrained optimization version 2.

    SIF input: Nick Gould, Nov 1992.

    Classification: QQR2-AN-V-V

    TODO: Human review needed - same constraint issues as EIGENA
    Related to systematic constraint value discrepancies in EIGENA implementation.
    This quadratic formulation likely has the same underlying
    constraint interpretation issues.
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem dimension
    N: int = 50  # Default from SIF file

    @property
    def n(self):
        """Number of variables: N eigenvalues + N*N eigenvector components."""
        return self.N + self.N * self.N

    @property
    def m(self):
        """Number of constraints: orthogonality constraints for upper triangular."""
        # Orthogonality equations Q^T Q - I = 0 for upper triangular part
        return self.N * (self.N + 1) // 2

    def objective(self, y, args):
        """Quadratic least squares objective function.

        The SIF uses L2 group type which creates a quadratic objective.
        Minimize ||Q^T D - A Q^T||_F^2 (Frobenius norm squared)
        """
        del args

        # Extract eigenvalues and eigenvectors
        d = y[: self.N]  # Eigenvalues D(j)
        q = y[self.N :].reshape(self.N, self.N)  # Eigenvectors Q(i,j)

        # Define matrix A: diagonal with eigenvalues 1, ..., N
        a = inexact_asarray(jnp.diag(jnp.arange(1, self.N + 1)))

        # Compute eigen-equation residuals: Q^T D - A Q^T = 0
        # Based on SIF: E(I,J) has elements Q1*Q2*D where Q1=Q(J,I), Q2=Q(K,J), D=D(K)
        # This suggests: sum_k Q(J,I) * Q(K,J) * D(K) which is (Q^T diag(D) Q)_{I,J}
        # But the SIF says "Q(T) D - A Q(T) = 0" so let's use that form
        qtd = q.T @ jnp.diag(d)  # Q^T D
        aqt = a @ q.T  # A Q^T

        residual_matrix = qtd - aqt

        # L2 objective: sum of squared residuals for all elements
        return inexact_asarray(jnp.sum(residual_matrix**2))

    def constraint(self, y):
        """Orthogonality constraints: Q^T Q - I = 0."""
        # Extract eigenvectors
        q = y[self.N :].reshape(self.N, self.N)

        # Compute Q^T Q - I
        qtq = q.T @ q
        eye = jnp.eye(self.N)
        ortho_residuals = qtq - eye

        # Extract upper triangular part as constraints
        constraints = []
        for j in range(self.N):
            for i in range(j + 1):
                constraints.append(ortho_residuals[i, j])

        return inexact_asarray(jnp.array(constraints)), None

    @property
    def y0(self):
        """Initial guess."""
        y0 = jnp.zeros(self.n)

        # From pycutest behavior (differs from SIF file):
        # Set D[0] and D[1] to 1.0
        y0 = y0.at[0].set(1.0)
        y0 = y0.at[1].set(1.0)

        # pycutest sets Q matrix with same pattern as EIGENA
        q_start = self.N
        for i in range(self.N):
            if i == 24:
                # Row 24 only has Q[24,25] = 1.0
                idx = q_start + i * self.N + 25
                y0 = y0.at[idx].set(1.0)
            elif i == 49:
                # Row 49 only has Q[49,49] = 1.0
                idx = q_start + i * self.N + 49
                y0 = y0.at[idx].set(1.0)
            elif i >= 25 and i < 49:
                # Rows 25-48 have a wrapped pattern
                # First column: 2*(i-25)+1
                col1 = 2 * (i - 25) + 1
                idx = q_start + i * self.N + col1
                y0 = y0.at[idx].set(1.0)
                # Second column: i+1
                col2 = i + 1
                idx = q_start + i * self.N + col2
                y0 = y0.at[idx].set(1.0)
            else:
                # Rows 0-23 follow the original pattern
                # Q[i, i+1] = 1.0
                if i + 1 < self.N:
                    idx = q_start + i * self.N + (i + 1)
                    y0 = y0.at[idx].set(1.0)

                # Q[i, 2*i+3] = 1.0
                if 2 * i + 3 < self.N:
                    idx = q_start + i * self.N + (2 * i + 3)
                    y0 = y0.at[idx].set(1.0)

        return inexact_asarray(y0)

    @property
    def args(self):
        """No additional arguments."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution: eigenvalues 1, ..., N and identity matrix."""
        # For diagonal matrix A with eigenvalues 1, ..., N,
        # the solution should satisfy A Q^T = Q^T D
        d_expected = inexact_asarray(jnp.arange(1, self.N + 1))
        q_expected = inexact_asarray(jnp.eye(self.N).flatten())
        return jnp.concatenate([d_expected, q_expected])

    @property
    def expected_objective_value(self):
        """Expected optimal objective value (0 for correct eigendecomposition)."""
        return inexact_asarray(jnp.array(0.0))

    @property
    def bounds(self) -> tuple[jnp.ndarray, jnp.ndarray] | None:
        """Free variables as specified in SIF file."""
        # FR EIGENA2 'DEFAULT' means all variables are free (no bounds)
        return None
